import asyncio
import logging

import kr8s.asyncio

logger = logging.getLogger("koreo.resource_function.reconcile")

LOOKUP_TIMEOUT = 15

_plural_map: dict[str, str] = {}

_lookup_locks: dict[str, asyncio.Event] = {}


async def get_plural_kind(
    api: kr8s.asyncio.Api, kind: str, api_version: str
) -> str | None:
    if api_version == "v1":
        lookup_kind = f"{kind}"
    else:
        lookup_kind = f"{kind}.{api_version}"

    if lookup_kind in _plural_map:
        return _plural_map[lookup_kind]

    lookup_lock = _lookup_locks.get(lookup_kind)
    if lookup_lock:
        try:
            await asyncio.wait_for(lookup_lock.wait(), timeout=LOOKUP_TIMEOUT)
        except asyncio.TimeoutError:
            pass

        if lookup_kind in _plural_map:
            return _plural_map[lookup_kind]

        raise Exception(f"Waiting on {lookup_kind} failed.")

    lookup_lock = asyncio.Event()
    _lookup_locks[lookup_kind] = lookup_lock

    for _ in range(3):
        try:
            async with asyncio.timeout(LOOKUP_TIMEOUT):
                try:
                    (_, plural_kind, _) = await api.lookup_kind(lookup_kind)
                    break
                except ValueError:
                    del _lookup_locks[lookup_kind]
                    logger.error(f"Failed to find Kind (`{lookup_kind}`) information.")
                    return None
        except asyncio.TimeoutError:
            continue
        except:
            del _lookup_locks[lookup_kind]
            raise
    else:
        del _lookup_locks[lookup_kind]
        raise Exception(
            f"Too many failed attempts to find plural kind for {lookup_kind} failed."
        )

    _plural_map[lookup_kind] = plural_kind

    lookup_lock.set()

    return plural_kind


def _reset():
    """Helper for unit testing; not intended for usage in normal code."""
    _plural_map.clear()

    for lock in _lookup_locks.values():
        lock.set()

    _lookup_locks.clear()



API_GROUP = PREFIX = "koreo.dev"

DEFAULT_API_VERSION = "v1beta1"

ACTIVE_LABEL = f"{PREFIX}/active"

LAST_APPLIED_ANNOTATION = f"{PREFIX}/last-applied-configuration"

DEFAULT_CREATE_DELAY = 30
DEFAULT_DELETE_DELAY = 15
DEFAULT_LOAD_RETRY_DELAY = 30
DEFAULT_PATCH_DELAY = 30

KOREO_DIRECTIVE_KEYS: set[str] = {
    "x-koreo-compare-as-set",
    "x-koreo-compare-as-map",
    "x-koreo-compare-last-applied",
}


PLURAL_LOOKUP_NEEDED = "+++missing plural+++"



from typing import Any

from koreo import registry
from koreo import result

from koreo.resource_function.structure import ResourceFunction
from koreo.value_function.structure import ValueFunction
from koreo.workflow.structure import Workflow

FunctionResource = (
    registry.Resource[ResourceFunction] | registry.Resource[ValueFunction]
)

LogicResource = (
    registry.Resource[ResourceFunction]
    | registry.Resource[ValueFunction]
    | registry.Resource[Workflow]
)


def function_ref_spec_to_resource(
    spec: Any, location: str
) -> None | FunctionResource | result.PermFail:
    if not spec:
        return None

    if not isinstance(spec, dict):
        return result.PermFail(
            message=(
                "Failed to process `functionRef`, expected object with `kind` and `name`. "
                f"received {type(spec)}"
            ),
            location=location,
        )

    logic_kind = spec.get("kind")
    logic_name = spec.get("name")

    if not logic_kind:
        return result.PermFail(message="Missing `functionRef.kind`.", location=location)

    if not logic_name:
        return result.PermFail(
            message=f"Missing `functionRef.name`.", location=location
        )

    # This is not using a "dict lookup" because of Python's deficient type
    # narrowing.
    match logic_kind:
        case "ValueFunction":
            return registry.Resource(resource_type=ValueFunction, name=logic_name)
        case "ResourceFunction":
            return registry.Resource(resource_type=ResourceFunction, name=logic_name)
        case _:
            return result.PermFail(
                message=f"Invalid `functionRef.kind` ({logic_kind}).", location=location
            )



import copy
import random
import string
import unittest

from celpy import celtypes

from koreo.resource_function.reconcile import validate

CHARS = string.ascii_letters + " "


def string_generator(size: int):
    return "".join(random.choices(population=CHARS, k=size))


class TestListValidator(unittest.TestCase):
    def test_match(self):
        generators = (
            ("strings", lambda: string_generator(random.randint(5, 25))),
            ("ints", lambda: random.randint(-500000000, 500000000)),
        )

        for label, generator in generators:
            objects = [generator() for _ in range(random.randint(5, 25))]

            lhs = objects[:]
            rhs = objects[:]

            match = validate._validate_list_match(target=lhs, actual=rhs)
            self.assertFalse(match.differences, f"{label} mismatch")
            self.assertTrue(match.match, f"{label} mismatch")

    def test_matching_lists_out_of_order(self):
        generators = (
            ("strings", lambda: string_generator(random.randint(5, 25))),
            ("ints", lambda: random.randint(-500000000, 500000000)),
        )

        for label, generator in generators:
            objects = [generator() for _ in range(random.randint(5, 25))]

            lhs = objects[:]

            random.shuffle(objects)

            rhs = objects[:]

            match = validate._validate_list_match(target=lhs, actual=rhs)
            self.assertTrue(match.differences, f"{label} mismatch")
            self.assertFalse(match.match, f"{label} mismatch")

    def test_general_mismatch(self):
        generators = (
            ("strings", lambda: string_generator(random.randint(5, 25))),
            ("ints", lambda: random.randint(-500000000, 500000000)),
        )

        for label, generator in generators:
            lhs = [generator() for _ in range(random.randint(0, 15))]

            rhs = [generator() for _ in range(random.randint(0, 15))]

            match = validate._validate_list_match(target=lhs, actual=rhs)
            self.assertTrue(match.differences, f"{label} mismatch")
            self.assertFalse(match.match, f"{label} mismatch")

    def test_extra_values(self):
        generators = (
            ("strings", lambda: string_generator(random.randint(5, 25))),
            ("ints", lambda: random.randint(-500000000, 500000000)),
        )

        for label, generator in generators:
            objects = [generator() for _ in range(random.randint(5, 25))]

            lhs = objects[:-3]
            rhs = objects[:]

            match = validate._validate_list_match(target=lhs, actual=rhs)
            self.assertEqual(1, len(match.differences), f"{label} mismatch")
            self.assertFalse(match.match, f"{label} mismatch")

            (difference_message, *_) = match.differences
            self.assertIn("length", difference_message, f"{label} mismatch")

    def test_nulls(self):
        non_nulls = [
            string_generator(random.randint(5, 50))
            for _ in range(random.randint(5, 50))
        ]

        match = validate._validate_list_match(target=non_nulls, actual=None)
        self.assertFalse(match.match)
        (difference_message, *_) = match.differences
        self.assertIn("null array", difference_message)

        match = validate._validate_list_match(target=None, actual=non_nulls)
        self.assertFalse(match.match)
        (difference_message, *_) = match.differences
        self.assertIn("null array", difference_message)

        match = validate._validate_list_match(target=None, actual=None)
        self.assertTrue(match.match)
        self.assertFalse(match.differences)

    def test_last_applied_matches_actual(self):
        generators = (
            ("strings", lambda: string_generator(random.randint(5, 25))),
            ("ints", lambda: random.randint(-500000000, 500000000)),
        )

        for label, generator in generators:
            objects = [generator() for _ in range(random.randint(5, 25))]

            lhs = objects[:]
            rhs = objects[:]
            last_applied = objects[:]

            match = validate._validate_list_match(
                target=lhs, actual=rhs, last_applied_value=last_applied
            )
            self.assertFalse(match.differences, f"{label} mismatch")
            self.assertTrue(match.match, f"{label} mismatch")

    def test_last_applied_too_short(self):
        generators = (
            ("strings", lambda: string_generator(random.randint(5, 25))),
            ("ints", lambda: random.randint(-500000000, 500000000)),
        )

        for label, generator in generators:
            objects = [generator() for _ in range(random.randint(5, 25))]

            lhs = objects[:]
            rhs = objects[:]
            last_applied = objects[:2]

            match = validate._validate_list_match(
                target=lhs, actual=rhs, last_applied_value=last_applied
            )
            self.assertFalse(match.differences, f"{label} mismatch")
            self.assertTrue(match.match, f"{label} mismatch")


class TestSetValidator(unittest.TestCase):
    def test_match(self):
        generators = (
            ("strings", lambda: string_generator(random.randint(5, 25))),
            ("ints", lambda: random.randint(-500000000, 500000000)),
        )

        for label, generator in generators:
            objects = [generator() for _ in range(random.randint(5, 25))]

            lhs = objects[:]
            random.shuffle(lhs)

            rhs = objects[:]
            random.shuffle(rhs)

            match = validate._validate_set_match(target=lhs, actual=rhs)
            self.assertFalse(match.differences, f"{label} mismatch")
            self.assertTrue(match.match, f"{label} mismatch")

    def test_partial_match(self):
        generators = (
            ("strings", lambda: string_generator(random.randint(5, 25))),
            ("ints", lambda: random.randint(-500000000, 500000000)),
        )

        for label, generator in generators:
            objects = [generator() for _ in range(random.randint(5, 25))]

            random.shuffle(objects)
            lhs = objects[: random.randint(3, len(objects) - 2)]

            random.shuffle(objects)
            rhs = objects[: random.randint(3, len(objects) - 2)]

            match = validate._validate_set_match(target=lhs, actual=rhs)
            self.assertTrue(match.differences, f"{label} mismatch")
            self.assertFalse(match.match, f"{label} mismatch")

    def test_extra_values(self):
        generators = (
            ("strings", lambda: string_generator(random.randint(5, 25))),
            ("ints", lambda: random.randint(-500000000, 500000000)),
        )

        for label, generator in generators:
            objects = [generator() for _ in range(random.randint(5, 25))]

            lhs = objects[:-3]

            rhs = objects[:]

            match = validate._validate_set_match(target=lhs, actual=rhs)
            self.assertEqual(1, len(match.differences), f"{label} mismatch")
            self.assertFalse(match.match, f"{label} mismatch")

            (difference_message, *_) = match.differences
            self.assertIn("unexpectedly", difference_message, f"{label} mismatch")

    def test_dict_error_message(self):
        objects = [
            {"name": string_generator(20)},
            {"name": string_generator(20)},
            {"name": string_generator(20)},
            {"name": string_generator(20)},
        ]

        lhs = objects[:]
        random.shuffle(lhs)

        rhs = objects[:]
        random.shuffle(rhs)

        match = validate._validate_set_match(target=lhs, actual=rhs)
        self.assertFalse(match.match)

        (difference_message, *_) = match.differences
        self.assertIn("x-koreo-compare-as-map", difference_message)

    def test_list_message(self):
        objects = [
            [
                string_generator(random.randint(5, 50))
                for _ in range(random.randint(5, 50))
            ],
            [
                string_generator(random.randint(5, 50))
                for _ in range(random.randint(5, 50))
            ],
        ]

        lhs = objects[:]
        random.shuffle(lhs)

        rhs = objects[:]
        random.shuffle(rhs)

        match = validate._validate_set_match(target=lhs, actual=rhs)
        self.assertFalse(match.match)

        (difference_message, *_) = match.differences
        self.assertIn("unhashable", difference_message)

    def test_nulls(self):
        non_nulls = [
            string_generator(random.randint(5, 50))
            for _ in range(random.randint(5, 50))
        ]

        match = validate._validate_set_match(target=non_nulls, actual=None)
        self.assertFalse(match.match)
        (difference_message, *_) = match.differences
        self.assertIn("null set", difference_message)

        match = validate._validate_set_match(target=None, actual=non_nulls)
        self.assertFalse(match.match)
        (difference_message, *_) = match.differences
        self.assertIn("null set", difference_message)

        match = validate._validate_set_match(target=None, actual=None)
        self.assertTrue(match.match)
        self.assertFalse(match.differences)


class TestDifferenceValidator(unittest.TestCase):
    def test_basics(self):
        cases = [
            ("", ""),
            (None, None),
            (True, True),
            (False, False),
            (0, 0),
            (1, 1),
            (10.2, 10.2),
            ([], []),
            ((), ()),
            ({}, {}),
        ]

        for target, actual in cases:
            match = validate.validate_match(target=target, actual=actual)
            self.assertFalse(match.differences)
            self.assertTrue(match.match)

    def test_type_mismatches(self):
        types = ["", None, False, True, [], {}, 0, 1, 8.2]

        cases = [
            (base, other)
            for base_idx, base in enumerate(types)
            for other_idx, other in enumerate(types)
            if base_idx != other_idx
        ]

        for target, actual in cases:
            match = validate.validate_match(target=target, actual=actual)
            self.assertTrue(
                match.differences,
                f"Failed to detect type-mismatch between '{target}' and '{actual}'",
            )
            self.assertFalse(match.match)

    def test_dict_match(self):
        target_list = {
            "one": 1,
            "two": False,
            "three": {
                "three.1": 0,
                "three.2": 1,
                "three.3": True,
                "three.4": False,
                "three.5": [1, 2, "a", "b", True, False, {"value": 1}],
                "three.6": {
                    "one": 1,
                    "a": "b",
                    "c": True,
                    "d": False,
                    "e": {"value": 1},
                },
            },
            "four": 7.2,
        }
        actual_list = copy.deepcopy(target_list)

        match = validate.validate_match(target=target_list, actual=actual_list)
        self.assertFalse(match.differences)
        self.assertTrue(match.match)

    def test_dict_mismatch(self):
        target_list = {
            "one": 1,
            "two": False,
            "three": {
                "three.1": 0,
                "three.2": 1,
                "three.3": True,
                "three.4": False,
                "three.5": [1, 2, "a", "b", True, False, {"value": 1}],
                "three.6": {
                    "one": 1,
                    "a": "b",
                    "c": True,
                    "d": False,
                    "e": {"value": 1},
                },
            },
            "four": 7.2,
        }
        actual_list = copy.deepcopy(target_list)
        actual_list["three"]["three.5"][6]["value"] = 2

        match = validate.validate_match(target=target_list, actual=actual_list)
        self.assertTrue(match.differences)
        self.assertIn("index '6'", "".join(match.differences))
        self.assertFalse(match.match)

    def test_dict_missing_key(self):
        target_list = {
            "one": 1,
            "two": False,
            "three": {"value": 1},
            "four": 7.2,
        }
        actual_list = copy.deepcopy(target_list)
        del actual_list["three"]

        match = validate.validate_match(target=target_list, actual=actual_list)
        self.assertTrue(match.differences)
        self.assertIn("missing", "".join(match.differences))
        self.assertFalse(match.match)

    def test_dict_owner_refs_skipped(self):
        target_list = {
            "one": 1,
            "two": False,
        }
        actual_list = copy.deepcopy(target_list)

        target_list["ownerReferences"] = {"one": "value"}
        actual_list["ownerReferences"] = {"one": "dfferent value"}

        match = validate.validate_match(target=target_list, actual=actual_list)
        self.assertFalse(match.differences)
        self.assertTrue(match.match)

    def test_list_match(self):
        target_list = [1, "two", {"three": 4, "five": "six"}, 7.2]
        actual_list = target_list.copy()

        match = validate.validate_match(target=target_list, actual=actual_list)
        self.assertFalse(match.differences)
        self.assertTrue(match.match)

    def test_list_length_mismatch(self):
        target_list = [1, "two", {"three": 4, "five": "six"}, 7.2]
        actual_list = [1, "two", {"three": 4, "five": 6}]

        match = validate.validate_match(target=target_list, actual=actual_list)
        self.assertTrue(match.differences)
        self.assertIn("length", "".join(match.differences))
        self.assertFalse(match.match)

    def test_list_mismatch(self):
        target_list = [1, "two", {"three": 4, "five": "six"}, 7.2]
        actual_list = [1, "two", {"three": 4, "five": 6}, 7.2]

        match = validate.validate_match(target=target_list, actual=actual_list)
        self.assertTrue(match.differences)
        self.assertFalse(match.match)

    def test_compare_as_map(self):
        target_list = {
            "one": 1,
            "two": False,
            "x-koreo-compare-as-map": {"as-map": ["case-name"]},
            "as-map": [
                {"case-name": "three.1", "value": 0},
                {"case-name": "three.2", "value": 1},
                {"case-name": "three.3", "value": True},
                {"case-name": "three.4", "value": False},
            ],
            "four": 7.2,
        }
        actual_list = copy.deepcopy(target_list)
        random.shuffle(actual_list["as-map"])

        match = validate.validate_match(target=target_list, actual=actual_list)
        self.assertFalse(match.differences)
        self.assertTrue(match.match)

    def test_compare_as_set(self):
        target_list = {
            "one": 1,
            "two": False,
            "x-koreo-compare-as-set": ["as-set"],
            "as-set": [
                string_generator(random.randint(5, 50))
                for _ in range(random.randint(5, 50))
            ],
            "four": 7.2,
        }
        actual_list = copy.deepcopy(target_list)
        random.shuffle(actual_list["as-set"])

        match = validate.validate_match(target=target_list, actual=actual_list)
        self.assertFalse(match.differences)
        self.assertTrue(match.match)

    def test_compare_last_applied_simple(self):
        target_value = {
            "one": 1,
            "two": False,
            "three": {
                "three.1": 0,
                "three.2": 1,
                "three.3": True,
                "three.4": False,
                "three.5": [1, 2, "a", "b", True, False, {"value": 1}],
                "three.6": {
                    "one": 1,
                    "a": "b",
                    "c": True,
                    "d": False,
                    "e": {"x-koreo-compare-last-applied": ["value"], "value": 1},
                },
            },
            "four": 7.2,
        }
        actual_value = copy.deepcopy(target_value)
        actual_value["three"]["three.6"]["e"]["value"] = 2

        last_applied_value = copy.deepcopy(target_value)

        match = validate.validate_match(
            target=target_value,
            actual=actual_value,
            last_applied_value=last_applied_value,
        )
        self.assertFalse(match.differences)
        self.assertTrue(match.match)

    def test_compare_last_applied_dict_diff(self):
        target_value = {
            "one": 1,
            "two": False,
            "x-koreo-compare-last-applied": ["to-last-applied"],
            "to-last-applied": {
                "three.1": 0,
                "three.2": 1,
                "three.3": True,
                "three.4": False,
                "three.5": [1, 2, "a", "b", True, False, {"value": 1}],
                "three.6": {
                    "one": 1,
                    "a": "b",
                    "c": True,
                    "d": False,
                    "e": {"value": 1},
                },
            },
            "four": 7.2,
        }
        actual_value = copy.deepcopy(target_value)
        actual_value["to-last-applied"] = {
            "totally": 0,
            "different": 1,
            "values": True,
            "go": ["into", "this", "structure"],
        }

        last_applied_value = copy.deepcopy(target_value)

        match = validate.validate_match(
            target=target_value,
            actual=actual_value,
            last_applied_value=last_applied_value,
        )
        self.assertFalse(match.differences)
        self.assertTrue(match.match)

    def test_compare_last_applied_list_diff(self):
        target_value = {
            "one": 1,
            "two": False,
            "x-koreo-compare-last-applied": ["to-last-applied"],
            "to-last-applied": ["abc", "one", 2, 3],
            "four": 7.2,
        }
        actual_value = copy.deepcopy(target_value)
        del actual_value["to-last-applied"]

        last_applied_value = copy.deepcopy(target_value)

        match = validate.validate_match(
            target=target_value,
            actual=actual_value,
            last_applied_value=last_applied_value,
        )
        self.assertFalse(match.differences)
        self.assertTrue(match.match)

    def test_compare_last_applied_set_diff(self):
        target_value = {
            "one": 1,
            "two": False,
            "x-koreo-compare-last-applied": ["to-last-applied"],
            "x-koreo-compare-as-set": ["to-last-applied"],
            "to-last-applied": ["abc", "one", 2, 3],
            "four": 7.2,
        }
        actual_value = copy.deepcopy(target_value)
        actual_value["to-last-applied"] = {"one": True}

        last_applied_value = copy.deepcopy(target_value)
        random.shuffle(last_applied_value["to-last-applied"])

        match = validate.validate_match(
            target=target_value,
            actual=actual_value,
            last_applied_value=last_applied_value,
        )
        self.assertFalse(match.differences)
        self.assertTrue(match.match)

    def test_compare_celstring_to_none(self):
        target_value = {
            "x-koreo-compare-last-applied": ["to-last-applied"],
            "to-last-applied": celtypes.StringType(string_generator(50)),
        }
        actual_value = copy.deepcopy(target_value)
        actual_value["to-last-applied"] = string_generator(20)

        last_applied_value = None

        match = validate.validate_match(
            target=target_value,
            actual=actual_value,
            last_applied_value=last_applied_value,
        )
        self.assertTrue(match.differences)
        self.assertFalse(match.match)



from typing import Sequence
import logging
import re

import celpy

from koreo import registry
from koreo import schema
from koreo.cache import get_resource_from_cache
from koreo.cel.functions import koreo_function_annotations
from koreo.cel.prepare import prepare_expression, prepare_map_expression
from koreo.cel.structure_extractor import extract_argument_structure
from koreo.resource_function.structure import ResourceFunction
from koreo.result import (
    Ok,
    Outcome,
    PermFail,
    Retry,
    UnwrappedOutcome,
    is_error,
    is_unwrapped_ok,
    unwrapped_combine,
)
from koreo.value_function.structure import ValueFunction

from . import structure

logger = logging.getLogger("koreo.workflow")


async def prepare_workflow(
    cache_key: str, spec: dict
) -> UnwrappedOutcome[tuple[structure.Workflow, Sequence[registry.Resource]]]:
    logger.debug(f"Prepare workflow {cache_key}")

    if error := schema.validate(
        resource_type=structure.Workflow, spec=spec, validation_required=True
    ):
        return PermFail(
            message=error.message,
            location=_location(cache_key, "spec"),
        )

    cel_env = celpy.Environment(annotations=koreo_function_annotations)

    # Used to  update our cross-reference registries
    watched_resources = set[registry.Resource]()

    steps_spec = spec.get("steps", [])
    step_resources, steps, steps_ready, parent_properties = _load_steps(
        cel_env, steps_spec
    )
    if step_resources:
        watched_resources.update(step_resources)

    if not steps_spec:
        steps_ready = unwrapped_combine(
            (
                steps_ready,
                PermFail(
                    message=(
                        "No steps specified Workflow at least one step in "
                        "'spec.steps' is required."
                    )
                ),
            )
        )

    crd_ref = _build_crd_ref(spec.get("crdRef", {}))

    return (
        structure.Workflow(
            name=cache_key,
            crd_ref=crd_ref,
            steps_ready=steps_ready,
            steps=steps,
            dynamic_input_keys=parent_properties,
        ),
        tuple(watched_resources),
    )


def _location(cache_key: str, extra: str | None = None) -> str:
    base = f"prepare:Workflow:{cache_key}"
    if not extra:
        return base

    return f"{base}:{extra}"


def _build_crd_ref(crd_ref_spec: dict) -> structure.ConfigCRDRef | None:
    api_group = crd_ref_spec.get("apiGroup")
    version = crd_ref_spec.get("version")
    kind = crd_ref_spec.get("kind")

    if not (api_group and version and kind):
        return None

    return structure.ConfigCRDRef(api_group=api_group, version=version, kind=kind)


PARENT_NAME_PATTERN = re.compile(r"parent.(?P<name>.*)")
STEPS_NAME_PATTERN = re.compile(r"steps.(?P<name>[^.[]+)?\[?.*")


LogicRegistryResource = registry.Resource[
    ValueFunction | ResourceFunction | structure.Workflow
]
Logic = ValueFunction | ResourceFunction | structure.Workflow


def _load_steps(
    cel_env: celpy.Environment,
    steps_spec: list[dict],
) -> tuple[
    set[LogicRegistryResource],
    Sequence[structure.Step | structure.ErrorStep],
    Outcome,
    set[str],
]:
    if not steps_spec:
        return set(), [], Ok(None), set()

    known_steps: set[str] = set()
    parent_properties: set[str] = set()

    resources: set[LogicRegistryResource] = set()
    step_outcomes: list[structure.Step | structure.ErrorStep] = []
    for step_spec in steps_spec:
        step_label: str = step_spec.get("label", "<missing label>")
        if step_label in known_steps:
            step_outcomes.append(
                structure.ErrorStep(
                    label=step_label,
                    outcome=PermFail(
                        message=f"Duplicate step-label ({step_label}) can not prepare Workflow."
                    ),
                    condition=None,
                )
            )
            continue
        step_resources, step_logic, needed_parent_properties = _load_step(
            cel_env, step_spec, known_steps
        )
        known_steps.add(step_label)
        parent_properties.update(needed_parent_properties)
        step_outcomes.append(step_logic)
        if step_resources:
            resources.update(step_resources)

    error_outcomes = [
        step.outcome for step in step_outcomes if isinstance(step, structure.ErrorStep)
    ]

    if not error_outcomes:
        return resources, step_outcomes, Ok(None), parent_properties

    return (
        resources,
        step_outcomes,
        unwrapped_combine(error_outcomes),
        parent_properties,
    )


def _load_step(
    cel_env: celpy.Environment, step_spec: dict, known_steps: set[str]
) -> tuple[
    set[LogicRegistryResource] | None, structure.Step | structure.ErrorStep, set[str]
]:
    step_label = step_spec.get("label", "<missing label>")

    step_location = f"spec.steps['{step_label}']"

    needed_steps = set()
    needed_parent_properties = set()

    resources = None
    logic = None

    logic_ref = step_spec.get("ref")
    logic_ref_switch = step_spec.get("refSwitch")
    if logic_ref and logic_ref_switch:
        # This should never happen due to schema validation
        return (
            None,
            structure.ErrorStep(
                label=step_label,
                outcome=PermFail(
                    message=f"spec.steps['{step_label}'] may not specify both `ref` and `refSwitch`, can not prepare Workflow.",
                    location=step_location,
                ),
                condition=None,
            ),
            needed_parent_properties,
        )

    if logic_ref:
        resources, logic = _load_logic(logic_ref=logic_ref, location=step_location)
    elif logic_ref_switch:
        resources, logic = _load_logic_switch(
            cel_env=cel_env, logic_switch=logic_ref_switch, location=step_location
        )
        if is_unwrapped_ok(logic):
            arg_structure = extract_argument_structure(logic.switch_on.ast)

            needed_parent_properties.update(
                match.group("name")
                for match in (PARENT_NAME_PATTERN.match(key) for key in arg_structure)
                if match
            )

            needed_steps.update(
                match.group("name")
                for match in (STEPS_NAME_PATTERN.match(key) for key in arg_structure)
                if match
            )

    if step_label == "<missing label>":
        # Note, this should be impossible due to schema validation.
        return (
            resources,
            structure.ErrorStep(
                label="missing",
                outcome=PermFail(
                    message="Missing step-label, can not prepare Workflow.",
                    location=step_location,
                ),
                condition=None,
            ),
            needed_parent_properties,
        )

    if not is_unwrapped_ok(logic):
        return (
            resources,
            structure.ErrorStep(label=step_label, outcome=logic, condition=None),
            needed_parent_properties,
        )

    if not logic:
        return (
            resources,
            structure.ErrorStep(
                label=step_label,
                outcome=PermFail(
                    message=f"Unable to load Logic for step ({step_label}), can not prepare Workflow.",
                    location=step_location,
                ),
                condition=None,
            ),
            needed_parent_properties,
        )

    skip_if_spec = step_spec.get("skipIf")
    match prepare_expression(
        cel_env=cel_env, spec=skip_if_spec, location=f"{step_location}.skipIf"
    ):
        case PermFail() as failure:
            return (
                resources,
                structure.ErrorStep(
                    label=step_label,
                    outcome=failure,
                    condition=None,
                ),
                needed_parent_properties,
            )
        case None:
            skip_if = None
        case celpy.Runner() as skip_if:
            arg_structure = extract_argument_structure(skip_if.ast)

            needed_parent_properties.update(
                match.group("name")
                for match in (PARENT_NAME_PATTERN.match(key) for key in arg_structure)
                if match
            )

            needed_steps.update(
                match.group("name")
                for match in (STEPS_NAME_PATTERN.match(key) for key in arg_structure)
                if match
            )

    for_each_spec = step_spec.get("forEach")
    match _prepare_for_each(cel_env=cel_env, step_label=step_label, spec=for_each_spec):
        case structure.ErrorStep() as error:
            return resources, error, needed_parent_properties
        case None:
            for_each = None
        case (structure.ForEach() as for_each, for_each_input_keys):
            needed_parent_properties.update(
                match.group("name")
                for match in (
                    PARENT_NAME_PATTERN.match(key) for key in for_each_input_keys
                )
                if match
            )

            needed_steps.update(
                match.group("name")
                for match in (
                    STEPS_NAME_PATTERN.match(key) for key in for_each_input_keys
                )
                if match
            )

    input_mapper_spec = step_spec.get("inputs")
    match prepare_map_expression(
        cel_env=cel_env, spec=input_mapper_spec, location=f"{step_location}.inputs"
    ):
        case PermFail() as failure:
            return (
                resources,
                structure.ErrorStep(
                    label=step_label,
                    outcome=failure,
                    condition=None,
                ),
                needed_parent_properties,
            )
        case None:
            input_mapper = None
        case celpy.Runner() as input_mapper:
            arg_structure = extract_argument_structure(input_mapper.ast)

            needed_parent_properties.update(
                match.group("name")
                for match in (PARENT_NAME_PATTERN.match(key) for key in arg_structure)
                if match
            )

            needed_steps.update(
                match.group("name")
                for match in (STEPS_NAME_PATTERN.match(key) for key in arg_structure)
                if match
            )

    condition_spec = step_spec.get("condition")
    if not condition_spec:
        condition = None
    else:
        condition = structure.StepConditionSpec(
            type_=condition_spec.get("type"),
            name=condition_spec.get("name"),
        )

    state_spec = step_spec.get("state")
    match prepare_map_expression(
        cel_env=cel_env, spec=state_spec, location=f"{step_location}.state"
    ):
        case None:
            state = None
        case celpy.Runner() as state:
            arg_structure = extract_argument_structure(state.ast)

            needed_parent_properties.update(
                match.group("name")
                for match in (PARENT_NAME_PATTERN.match(key) for key in arg_structure)
                if match
            )

            needed_steps.update(
                match.group("name")
                for match in (STEPS_NAME_PATTERN.match(key) for key in arg_structure)
                if match
            )
        case PermFail() as failure:
            return (
                resources,
                structure.ErrorStep(
                    label=step_label,
                    outcome=failure,
                    condition=None,
                ),
                needed_parent_properties,
            )

    out_of_order_steps = needed_steps.difference(known_steps)
    if out_of_order_steps:
        return (
            resources,
            structure.ErrorStep(
                label=step_label,
                outcome=PermFail(
                    message=f"Step '{step_label}', must come after {', '.join(out_of_order_steps)}.",
                    location=step_location,
                ),
                condition=None,
            ),
            needed_parent_properties,
        )

    return (
        resources,
        structure.Step(
            label=step_label,
            logic=logic,
            skip_if=skip_if,
            for_each=for_each,
            inputs=input_mapper,
            dynamic_input_keys=needed_steps,
            condition=condition,
            state=state,
        ),
        needed_parent_properties,
    )


def _prepare_for_each(
    cel_env: celpy.Environment,
    step_label: str,
    spec: dict | None,
) -> None | structure.ErrorStep | tuple[structure.ForEach, set[str]]:
    if not spec:
        return None

    source_iterator_spec = spec.get("itemIn")
    match prepare_expression(
        cel_env=cel_env,
        spec=source_iterator_spec,
        location=f"{step_label}.forEach.itemIn",
    ):
        case None:
            return structure.ErrorStep(
                label=step_label,
                outcome=PermFail(
                    message=f"Empty `{step_label}.forEach.itemIn`, can not prepare Workflow."
                ),
                condition=None,
            )
        case PermFail(message=message):
            return structure.ErrorStep(
                label=step_label,
                outcome=PermFail(
                    message=f"Error preparing `{step_label}.forEach.itemIn`: {message}."
                ),
                condition=None,
            )
        case celpy.Runner() as source_iterator:
            dynamic_input_keys = extract_argument_structure(source_iterator.ast)

    input_key = spec.get("inputKey")
    if not input_key:
        return structure.ErrorStep(
            label=step_label,
            outcome=PermFail(
                message=f"Missing `{step_label}.forEach.inputKey`, can not prepare Workflow."
            ),
            condition=None,
        )

    condition_spec = spec.get("condition")
    if not condition_spec:
        condition = None
    else:
        condition = structure.StepConditionSpec(
            type_=condition_spec.get("type"),
            name=condition_spec.get("name"),
        )

    for_each = structure.ForEach(
        source_iterator=source_iterator,
        input_key=input_key,
        condition=condition,
    )

    return (for_each, dynamic_input_keys)


def _load_logic(
    logic_ref: dict, location: str
) -> tuple[set[LogicRegistryResource] | None, PermFail | Retry | Logic]:
    logic_kind: str | None = logic_ref.get("kind")
    logic_cache_key: str | None = logic_ref.get("name")

    if not logic_kind:
        return (
            None,
            PermFail(
                message=f"Missing `{location}.kind`, can not prepare Workflow.",
                location=f"{location}.<missing>:{logic_cache_key if logic_cache_key else '<missing>'}",
            ),
        )

    if not logic_cache_key:
        return (
            None,
            PermFail(
                message=f"Missing `{location}.name`, can not prepare Workflow.",
                location=f"{location}.{logic_kind}:<missing>",
            ),
        )

    logic_kind_map: dict[
        str, type[ValueFunction | ResourceFunction | structure.Workflow]
    ] = {
        "ValueFunction": ValueFunction,
        "ResourceFunction": ResourceFunction,
        "Workflow": structure.Workflow,
    }

    logic_class = logic_kind_map.get(logic_kind)
    if not logic_class:
        return (
            None,
            PermFail(
                message=f"Invalid `{location}.kind` ({logic_kind}), can not prepare Workflow.",
                location=f"{location}.{logic_kind}:{logic_cache_key}",
            ),
        )

    resources = set(
        [
            registry.Resource(
                resource_type=logic_class,
                name=logic_cache_key,
            )
        ]
    )
    logic = get_resource_from_cache(
        resource_class=logic_class,
        cache_key=logic_cache_key,
    )

    if not logic:
        return (
            resources,
            Retry(
                message=f"'{logic_kind}:{logic_cache_key}' not cached; will retry.",
                delay=15,
                location=f"{location}.{logic_kind}:{logic_cache_key}",
            ),
        )

    if not is_unwrapped_ok(logic):
        return (
            resources,
            Retry(
                message=f"'{logic_kind}:{logic_cache_key}' is not healthy ({logic.message}); will retry.",
                delay=180,
                location=f"{location}.{logic_kind}:{logic_cache_key}",
            ),
        )

    return resources, logic


def _load_logic_switch(
    cel_env: celpy.Environment, logic_switch: dict, location: str
) -> tuple[set[LogicRegistryResource] | None, PermFail | Retry | structure.LogicSwitch]:
    dynamic_input_keys: set[str] = set()

    switch_on_spec = logic_switch.get("switchOn")
    match prepare_expression(cel_env=cel_env, spec=switch_on_spec, location=location):
        case None:
            return None, PermFail(
                message=f"Missing `{location}.switchOn`, can not prepare Workflow.",
                location=f"{location}.switchOn",
            )
        case PermFail() as err:
            return None, err
        case celpy.Runner() as switch_on:
            dynamic_input_keys.update(extract_argument_structure(switch_on.ast))

    cases_spec = logic_switch.get("cases")
    if not cases_spec:
        return None, PermFail(
            message=f"Must specify at least one case in `{location}.cases`, can not prepare Workflow.",
            location=f"{location}.cases",
        )

    default_logic = None
    logic_map = {}
    resources = set[LogicRegistryResource]()
    for idx, case_spec in enumerate(cases_spec):
        case = case_spec.get("case")
        is_default = case_spec.get("default")
        if is_default and default_logic:
            return None, PermFail(
                message=f"Only one case in `{location}.cases` may be default, can not prepare Workflow.",
                location=f"{location}.cases[{idx}].default",
            )

        logic_resources, logic = _load_logic(
            logic_ref=case_spec, location=f"{location}[{idx}]"
        )

        logic_map[case] = logic

        if is_unwrapped_ok(logic):
            if isinstance(logic, structure.Workflow):
                dynamic_input_keys.update(
                    f"inputs.{input}" for input in logic.dynamic_input_keys
                )
            else:
                dynamic_input_keys.update(logic.dynamic_input_keys)

        if is_default:
            default_logic = logic

        if logic_resources:
            resources.update(logic_resources)

    functions_ready = unwrapped_combine(logic_map.values())
    if is_error(functions_ready):
        return resources, functions_ready

    # This is just for the type checker
    if not is_unwrapped_ok(default_logic):
        return resources, default_logic

    return resources, structure.LogicSwitch(
        switch_on=switch_on,
        logic_map=logic_map,
        default_logic=default_logic,
        dynamic_input_keys=dynamic_input_keys,
    )



import copy
import logging

logger = logging.getLogger("koreo.cel.evaluation")

from celpy import celtypes
from celpy.celparser import tree_dump
import celpy

from koreo.predicate_helpers import predicate_to_koreo_result
from koreo.result import NonOkOutcome, PermFail

from koreo.cel.prepare import Index, Overlay


def evaluate(
    expression: celpy.Runner | None, inputs: dict[str, celtypes.Value], location: str
) -> None | celtypes.Value | PermFail:
    if not expression:
        return None

    try:
        expression_value = expression.evaluate(inputs)

        if eval_errors := check_for_celevalerror(expression_value, location):
            return eval_errors

        return expression_value

    except celpy.CELEvalError as err:
        tree = tree_dump(err.tree) if err and err.tree else ""
        return PermFail(
            message=f"Error evaluating `{location}` (at {tree}) {err.args}",
            location=tree,
        )

    except:
        msg = f"Unknown failure evaluating `{location}`."
        logger.exception(msg)
        return PermFail(msg)


def evaluate_predicates(
    predicates: celpy.Runner | None, inputs: dict[str, celtypes.Value], location: str
) -> None | NonOkOutcome:
    if not predicates:
        return None

    try:
        raw_result = predicates.evaluate(inputs)
        if eval_error := check_for_celevalerror(raw_result, location):
            return eval_error

        # This should be impossible, unless prepare validation was missed.
        if not isinstance(raw_result, celtypes.ListType):
            return PermFail(
                f"Bad structure for `{location}`, expected list of assertions but received {type(raw_result)}.",
                location=location,
            )

        return predicate_to_koreo_result(raw_result, location=location)

    except celpy.CELEvalError as err:
        tree = tree_dump(err.tree) if err and err.tree else ""
        return PermFail(
            message=f"Error evaluating `{location}` (at {tree}) {err.args}",
            location=tree,
        )
    except Exception as err:
        return PermFail(f"Error evaluating `{location}`: {err}", location=location)


def evaluate_overlay(
    overlay: Overlay,
    inputs: dict[str, celtypes.Value],
    base: celtypes.MapType,
    location: str,
) -> PermFail | celtypes.MapType:
    if not isinstance(overlay.value_index, dict):
        return PermFail(
            f"Bad overlay structure for `{location}`, expected mapping",
            location=location,
        )

    combined_inputs = inputs | {celtypes.StringType("resource"): base}

    try:
        overlay_values = overlay.values.evaluate(combined_inputs)
        if eval_error := check_for_celevalerror(overlay_values, location):
            return eval_error

        if not isinstance(overlay_values, celtypes.ListType):
            return PermFail(
                f"Bad overlay structure for `{location}`, received {type(overlay_values)}.",
                location=location,
            )

    except celpy.CELEvalError as err:
        tree = tree_dump(err.tree) if err and err.tree else ""
        return PermFail(
            message=f"Error evaluating `{location}` (at {tree}) {err.args}",
            location=tree,
        )

    except:
        msg = f"Unknown failure evaluating `{location}`."
        logger.exception(msg)
        return PermFail(msg)

    return _overlay_applier(base, index=overlay.value_index, values=overlay_values)


def _overlay_applier(
    base: celtypes.MapType, index: dict[str, Index], values: celtypes.ListType
) -> celtypes.MapType:
    overlaid = copy.deepcopy(base)
    for key, value_index in index.items():
        cel_key = celtypes.StringType(key)
        match value_index:
            case int():
                overlaid[cel_key] = values[value_index]
            case dict():
                match base.get(cel_key):
                    case dict() as sub_overlaid:
                        overlaid[cel_key] = _overlay_applier(
                            base=sub_overlaid,
                            index=value_index,
                            values=values,
                        )
                    case None | _:
                        overlaid[cel_key] = _overlay_applier(
                            base=celtypes.MapType(),
                            index=value_index,
                            values=values,
                        )
    return overlaid


def check_for_celevalerror(
    value: celtypes.Value | celpy.CELEvalError, location: str | None
) -> None | PermFail:
    match value:
        case celpy.CELEvalError(tree=error_tree):
            tree = tree_dump(error_tree) if error_tree else ""
            return PermFail(
                message=f"Error evaluating `{location}` (at {tree}) {value.args}",
                location=tree,
            )

        case celtypes.MapType() | dict():
            for key, subvalue in value.items():
                if eval_error := check_for_celevalerror(key, location):
                    return eval_error

                if eval_error := check_for_celevalerror(subvalue, location):
                    return eval_error

        case celtypes.ListType() | list() | tuple():
            for subvalue in value:
                if eval_error := check_for_celevalerror(subvalue, location):
                    return eval_error

    return None



from typing import Sequence
import logging
import re

logger = logging.getLogger("koreo.valuefunction.prepare")

from kr8s._objects import APIObject
import kr8s

import celpy

from koreo import cache
from koreo import constants
from koreo import registry
from koreo import schema
from koreo.cel.functions import koreo_function_annotations
from koreo.cel.prepare import (
    Overlay,
    prepare_expression,
    prepare_map_expression,
    prepare_overlay_expression,
)
from koreo.cel.structure_extractor import extract_argument_structure
from koreo.predicate_helpers import predicate_extractor
from koreo.result import (
    PermFail,
    Retry,
    UnwrappedOutcome,
    is_unwrapped_ok,
    unwrapped_combine,
)
from koreo.value_function.structure import ValueFunction

from . import structure

# Try to reduce the incredibly verbose logging from celpy
logging.getLogger("Environment").setLevel(logging.WARNING)
logging.getLogger("NameContainer").setLevel(logging.WARNING)
logging.getLogger("Evaluator").setLevel(logging.WARNING)
logging.getLogger("evaluation").setLevel(logging.WARNING)
logging.getLogger("celtypes").setLevel(logging.WARNING)


async def prepare_resource_function(
    cache_key: str, spec: dict
) -> UnwrappedOutcome[tuple[structure.ResourceFunction, set[registry.Resource]]]:
    logger.debug(f"Prepare ResourceFunction:{cache_key}")

    if error := schema.validate(
        resource_type=structure.ResourceFunction, spec=spec, validation_required=True
    ):
        return PermFail(
            error.message,
            location=_location(cache_key, "spec"),
        )

    env = celpy.Environment(annotations=koreo_function_annotations)

    used_vars = set[str]()

    match predicate_extractor(cel_env=env, predicate_spec=spec.get("preconditions")):
        case PermFail(message=message):
            return PermFail(
                message=message, location=_location(cache_key, "spec.preconditions")
            )
        case None:
            preconditions = None
        case celpy.Runner() as preconditions:
            used_vars.update(extract_argument_structure(preconditions.ast))

    match prepare_map_expression(
        cel_env=env, spec=spec.get("locals"), location="spec.locals"
    ):
        case PermFail(message=message):
            return PermFail(
                message=message,
                location=_location(cache_key, "spec.locals"),
            )
        case None:
            local_values = None
        case celpy.Runner() as local_values:
            used_vars.update(extract_argument_structure(local_values.ast))

    match _prepare_api_config(cel_env=env, spec=spec.get("apiConfig")):
        case PermFail(message=message):
            return PermFail(
                message=message,
                location=_location(cache_key, "spec.apiConfig"),
            )
        case (resource_api, resource_id, own_resource, readonly, delete_if_exists):
            used_vars.update(extract_argument_structure(resource_id.ast))

    match _prepare_resource_template(cel_env=env, spec=spec):
        case PermFail() as err:
            return err

        case structure.ResourceTemplateRef(name=template_name) as resource_template:
            if template_name:
                used_vars.update(extract_argument_structure(template_name.ast))

        case structure.InlineResourceTemplate(template=template) as resource_template:
            if template:
                used_vars.update(extract_argument_structure(template.ast))

    match _prepare_overlays(cel_env=env, spec=spec.get("overlays")):
        case None:
            overlays = []
            used_value_functions: set[registry.Resource] = set()

        case (overlays, overlay_input_keys, used_value_functions):
            used_vars.update(overlay_input_keys)

    match _prepare_create(cel_env=env, spec=spec.get("create")):
        case PermFail(message=message):
            return PermFail(
                message=message, location=_location(cache_key, "spec.create")
            )
        case structure.Create(overlay=overlay) as create:
            if overlay:
                used_vars.update(extract_argument_structure(overlay.values.ast))

    match _prepare_update(spec=spec.get("update")):
        case PermFail(message=message):
            return PermFail(
                message=message, location=_location(cache_key, "spec.update")
            )
        case (_) as update:
            # We just needed `update` set.
            pass

    match predicate_extractor(cel_env=env, predicate_spec=spec.get("postconditions")):
        case PermFail(message=message):
            return PermFail(
                message=message, location=_location(cache_key, "spec.postconditions")
            )
        case None:
            postconditions = None
        case celpy.Runner() as postconditions:
            used_vars.update(extract_argument_structure(postconditions.ast))

    match prepare_map_expression(
        cel_env=env, spec=spec.get("return"), location="spec.return"
    ):
        case PermFail(message=message) as err:
            return err
        case None:
            return_value = None
        case celpy.Runner() as return_value:
            used_vars.update(extract_argument_structure(return_value.ast))

    return (
        structure.ResourceFunction(
            name=cache_key,
            preconditions=preconditions,
            local_values=local_values,
            crud_config=structure.CRUDConfig(
                resource_api=resource_api,
                resource_id=resource_id,
                own_resource=own_resource,
                readonly=readonly,
                delete_if_exists=delete_if_exists,
                resource_template=resource_template,
                overlays=overlays,
                create=create,
                update=update,
            ),
            postconditions=postconditions,
            return_value=return_value,
            dynamic_input_keys=used_vars,
        ),
        used_value_functions,
    )


def _location(cache_key: str, extra: str | None = None) -> str:
    base = f"prepare:ResourceFunction:{cache_key}"
    if not extra:
        return base

    return f"{base}:{extra}"


def _prepare_api_config(
    cel_env: celpy.Environment, spec: dict
) -> tuple[type[APIObject], celpy.Runner, bool, bool, bool] | PermFail:
    api_version = spec.get("apiVersion")
    kind = spec.get("kind")
    name = spec.get("name")
    if not (api_version and kind and name):
        return PermFail(
            message="`apiVersion`, `kind`, and `name` are required in `spec.apiConfig`"
        )

    plural = spec.get("plural")
    if not plural:
        plural = constants.PLURAL_LOOKUP_NEEDED

    namespaced = spec.get("namespaced", True)
    owned = spec.get("owned", True)
    readonly = spec.get("readonly", False)
    delete_if_exists = spec.get("deleteIfExists", False)

    resource_id_cel = {"name": name}

    namespace = spec.get("namespace")
    if namespace:
        resource_id_cel["namespace"] = namespace
    elif namespaced:
        return PermFail(
            message="`namespace` is required when `spec.apiConfig.namespaced` is `true`"
        )

    match prepare_map_expression(
        cel_env=cel_env, spec=resource_id_cel, location="spec.apiConfig"
    ):
        case PermFail(message=message):
            return PermFail(message=message)
        case None:
            return PermFail(message="Error processing `spec.apiConfig.name`")
        case celpy.Runner() as resource_id:
            # Just needed to set name
            pass

    try:
        resource_api = kr8s.objects.get_class(
            version=api_version,
            kind=kind,
            _asyncio=True,
        )
    except KeyError:
        resource_api = kr8s.objects.new_class(
            version=api_version,
            kind=kind,
            plural=plural,
            namespaced=namespaced,
            asyncio=True,
        )
    return (resource_api, resource_id, owned, readonly, delete_if_exists)


def _prepare_resource_template(
    cel_env: celpy.Environment, spec: dict
) -> structure.InlineResourceTemplate | structure.ResourceTemplateRef | PermFail:
    match spec:
        case {"resource": resource_template}:
            match prepare_map_expression(
                cel_env=cel_env, spec=resource_template, location="spec.resource"
            ):
                case PermFail() as err:
                    return err
                case None:
                    return structure.InlineResourceTemplate()
                case celpy.Runner() as template_expression:
                    return structure.InlineResourceTemplate(
                        template=template_expression
                    )

        case {"resourceTemplateRef": resource_template_ref}:
            name_cel = resource_template_ref.get("name")
            match prepare_expression(
                cel_env=cel_env, spec=name_cel, location="spec.resourceTemplateRef.name"
            ):
                case None:
                    return PermFail(
                        message=f"`name` is required in `spec.resourceTemplateRef`",
                        location="spec.resourceTemplateRef.name",
                    )
                case PermFail() as err:
                    return err
                case celpy.Runner() as name_expression:
                    return structure.ResourceTemplateRef(name=name_expression)

        case _:
            return PermFail(
                message="One of `resource` or `resourceTemplateRef` is required.",
                location="spec",
            )


INPUTS_NAME_PATTERN = re.compile(r"inputs.(?P<name>[^.[]+)?\[?.*")


def _prepare_overlays(
    cel_env: celpy.Environment, spec: list[dict] | None
) -> (
    None
    | tuple[
        UnwrappedOutcome[
            Sequence[structure.ValueFunctionOverlay | structure.InlineOverlay]
        ],
        set[str],
        set[registry.Resource[ValueFunction]],
    ]
):
    if not spec:
        return None

    overlays = []
    used_inputs: set[str] = set()
    used_value_functions: set[registry.Resource[ValueFunction]] = set()

    for idx, overlay_spec in enumerate(spec):
        skip_if_spec = overlay_spec.pop("skipIf", None)

        match prepare_expression(
            cel_env=cel_env,
            spec=skip_if_spec,
            location=f"spec.overlays[{idx}].skipIf",
        ):
            case PermFail() as err:
                overlays.append(err)
                continue
            case None:
                skip_if = None
            case celpy.Runner() as skip_if:
                used_inputs.update(extract_argument_structure(skip_if.ast))

        match overlay_spec:
            case {"overlay": inline_spec, **extras}:
                if extras:
                    extra_keys = ", ".join(f"{key}" for key in extras.keys())
                    overlays.append(
                        PermFail(
                            message=f"may not specify {extra_keys} with overlay",
                            location=f"spec.overlays[{idx}].overlay",
                        )
                    )
                    continue

                match prepare_overlay_expression(
                    cel_env=cel_env,
                    spec=inline_spec,
                    location=f"spec.overlays[{idx}].overlay",
                ):
                    case None:
                        overlays.append(
                            PermFail(
                                message=f"Empty inline overlay ({overlay_spec})",
                                location=f"spec.overlays[{idx}].overlay",
                            )
                        )
                        continue
                    case PermFail() as err:
                        overlays.append(err)
                        continue
                    case Overlay() as overlay:
                        overlays.append(
                            structure.InlineOverlay(overlay=overlay, skip_if=skip_if)
                        )
                        used_inputs.update(
                            extract_argument_structure(overlay.values.ast)
                        )

            case {"overlayRef": {"name": overlay_name}, **extras}:
                used_value_functions.add(
                    registry.Resource(resource_type=ValueFunction, name=overlay_name)
                )

                inputs_spec = extras.pop("inputs", None)
                if extras:
                    extra_keys = ", ".join(f"{key}" for key in extras.keys())
                    overlays.append(
                        PermFail(
                            message=f"may not specify {extra_keys} with overlayRef",
                            location=f"spec.overlays[{idx}].overlayRef",
                        )
                    )
                    continue

                match prepare_map_expression(
                    cel_env=cel_env,
                    spec=inputs_spec,
                    location=f"spec.overlays[{idx}].inputs",
                ):
                    case PermFail() as err:
                        overlays.append(err)
                        continue
                    case None:
                        inputs = None
                    case celpy.Runner() as inputs:
                        used_inputs.update(extract_argument_structure(inputs.ast))

                overlay_resource = cache.get_resource_from_cache(
                    resource_class=ValueFunction, cache_key=overlay_name
                )
                if not is_unwrapped_ok(overlay_resource) or overlay_resource is None:
                    overlays.append(
                        Retry(
                            message=f"ValueFunction:{overlay_name} not ready ({overlay_resource})",
                            location=f"spec.overlays[{idx}].overlayRef",
                        )
                    )
                    continue

                provided_inputs: set[str] = set(
                    inputs_spec.keys() if inputs_spec else ()
                )
                needed_inputs: set[str] = {
                    match.group("name")
                    for match in (
                        INPUTS_NAME_PATTERN.match(key)
                        for key in overlay_resource.dynamic_input_keys
                    )
                    if match
                }
                missing_inputs = needed_inputs - provided_inputs
                if missing_inputs:
                    missing_input_keys = ", ".join(
                        f'"{missing}"' for missing in missing_inputs
                    )
                    overlays.append(
                        PermFail(
                            message=f"{overlay_name} expected the following inputs {missing_input_keys}.",
                            location=f"spec.overlays[{idx}].inputs",
                        )
                    )
                    continue

                overlays.append(
                    structure.ValueFunctionOverlay(
                        overlay=overlay_resource, skip_if=skip_if, inputs=inputs
                    )
                )

            case _:
                overlays.append(
                    PermFail(
                        message=f"Invalid overlays spec: {overlay_spec}",
                        location=f"spec.overlays[{idx}]",
                    )
                )

    return (unwrapped_combine(overlays), used_inputs, used_value_functions)


def _prepare_create(
    cel_env: celpy.Environment, spec: dict | None
) -> structure.Create | PermFail:
    if not spec:
        return structure.Create(enabled=True, delay=constants.DEFAULT_CREATE_DELAY)

    enabled = spec.get("enabled", True)
    if not enabled:
        return structure.Create(enabled=False)

    delay = spec.get("delay", constants.DEFAULT_CREATE_DELAY)

    overlay_spec = spec.get("overlay")
    match prepare_overlay_expression(
        cel_env=cel_env,
        spec=overlay_spec,
        location="spec.create.overlay",
    ):
        case PermFail() as err:
            return err
        case None:
            overlay = None
        case Overlay() as overlay:
            # Just needed overlay
            pass

    return structure.Create(enabled=enabled, delay=delay, overlay=overlay)


def _prepare_update(spec: dict | None) -> structure.Update | PermFail:
    match spec:
        case None:
            return structure.UpdatePatch(delay=constants.DEFAULT_PATCH_DELAY)
        case {"patch": {"delay": delay}}:
            return structure.UpdatePatch(delay=delay)
        case {"recreate": {"delay": delay}}:
            return structure.UpdateRecreate(delay=delay)
        case {"never": {}}:
            return structure.UpdateNever()
        case _:
            return PermFail(
                message="Malformed `spec.update`, expected a mapping with `patch`, `recreate`, or `never`"
            )



from typing import NamedTuple, Sequence

from kr8s._objects import APIObject

import celpy

from koreo.cel.prepare import Overlay
from koreo.result import UnwrappedOutcome
from koreo.value_function.structure import ValueFunction


class ResourceTemplateRef(NamedTuple):
    name: celpy.Runner | None


class InlineResourceTemplate(NamedTuple):
    template: celpy.Runner | None = None


class ValueFunctionOverlay(NamedTuple):
    overlay: ValueFunction
    skip_if: celpy.Runner | None = None
    inputs: celpy.Runner | None = None


class InlineOverlay(NamedTuple):
    overlay: Overlay
    skip_if: celpy.Runner | None = None


class Create(NamedTuple):
    enabled: bool = True
    delay: int = 30
    overlay: Overlay | None = None


class UpdatePatch(NamedTuple):
    delay: int = 30


class UpdateRecreate(NamedTuple):
    delay: int = 30


class UpdateNever(NamedTuple):
    pass


Update = UpdatePatch | UpdateRecreate | UpdateNever


class DeleteAbandon(NamedTuple):
    pass


class DeleteDestroy(NamedTuple):
    force: bool = False


class CRUDConfig(NamedTuple):
    resource_api: type[APIObject]
    resource_id: celpy.Runner
    own_resource: bool
    readonly: bool
    delete_if_exists: bool

    resource_template: InlineResourceTemplate | ResourceTemplateRef
    overlays: UnwrappedOutcome[Sequence[InlineOverlay | ValueFunctionOverlay]]

    create: Create
    update: Update


class ResourceFunction(NamedTuple):
    name: str
    preconditions: celpy.Runner | None
    local_values: celpy.Runner | None

    crud_config: CRUDConfig

    postconditions: celpy.Runner | None
    return_value: celpy.Runner | None

    dynamic_input_keys: set[str]



from typing import NamedTuple, Sequence

from celpy import celtypes

from koreo.cel.prepare import Overlay
from koreo.resource_function.structure import ResourceFunction
from koreo.result import NonOkOutcome, UnwrappedOutcome
from koreo.value_function.structure import ValueFunction


class ExpectOutcome(NamedTuple):
    outcome: NonOkOutcome | None = None


class ExpectReturn(NamedTuple):
    value: dict | None = None


class ExpectResource(NamedTuple):
    resource: dict | None = None


class ExpectDelete(NamedTuple):
    expected: bool = False


Assertion = ExpectOutcome | ExpectResource | ExpectReturn | ExpectDelete


class TestCase(NamedTuple):
    assertion: Assertion

    # Variant test cases do not carry forward.
    variant: bool = False

    # Skip
    skip: bool = False

    # For non-vaiant cases, these updates to inputs carry forward.
    input_overrides: celtypes.MapType | None = None

    # Mutually exclusive
    current_resource: dict | None = None
    resource_overlay: Overlay | None = None

    # Human friendly output
    label: str | None = None


class FunctionTest(NamedTuple):
    function_under_test: UnwrappedOutcome[ResourceFunction | ValueFunction]

    inputs: celtypes.MapType | None
    initial_resource: dict | None

    test_cases: Sequence[TestCase] | None



from typing import NamedTuple
import time

from koreo import cache
from koreo import registry
from koreo.result import is_ok, is_unwrapped_ok
from koreo.workflow.structure import Workflow


class Status(NamedTuple):
    resource_type: str
    cache_key: str
    resource_version: str
    prepared_ago_seconds: str
    resource_status: str
    subscriptions: list
    subscribers: list


def _get_resource_status_for_type(resource):
    if not is_unwrapped_ok(resource):
        return f"{resource}"

    match resource:
        case Workflow(steps_ready=steps_ready):
            if is_ok(steps_ready):
                return "Ok"

            return f"Steps not ready due to {steps_ready}"

        case _:
            return "Ok"


def list_resources():
    request_time = time.monotonic()

    cached_resources: list[Status] = []
    for resource_key, cached_resource in cache.__CACHE.items():
        subscribers = registry.get_subscribers(resource_key)
        subscriptions = registry.get_subscriptions(resource_key)

        prepared_ago_seconds = "never"
        if cached_resource.prepared_at:
            age = request_time - cached_resource.prepared_at
            prepared_ago_seconds = f"{age:.4f}"

        cached_resources.append(
            Status(
                resource_type=resource_key.resource_type.__qualname__,
                cache_key=resource_key.name,
                resource_version=cached_resource.resource_version,
                prepared_ago_seconds=prepared_ago_seconds,
                resource_status=_get_resource_status_for_type(cached_resource.resource),
                subscriptions=list(subscriptions),
                subscribers=list(subscribers),
            )
        )

    return cached_resources



import copy
import unittest

import celpy

from koreo import result

from koreo.value_function import prepare
from koreo.value_function import reconcile
from koreo.value_function.structure import ValueFunction


class TestReconcileValueFunction(unittest.IsolatedAsyncioTestCase):

    async def test_no_spec_returns_none(self):
        result = await reconcile.reconcile_value_function(
            location="test-fn",
            function=ValueFunction(
                preconditions=None,
                local_values=None,
                return_value=None,
                dynamic_input_keys=set(),
            ),
            inputs=celpy.json_to_cel({}),
        )

        self.assertEqual(result, None)

    async def test_full_scenario(self):
        prepared = await prepare.prepare_value_function(
            cache_key="test",
            spec={
                "preconditions": [
                    {
                        "assert": "=!inputs.preconditions.skip",
                        "skip": {"message": "=inputs.messages.skip"},
                    },
                    {
                        "assert": "=!inputs.preconditions.permFail",
                        "permFail": {"message": "=inputs.messages.permFail"},
                    },
                    {
                        "assert": "=!inputs.preconditions.depSkip",
                        "depSkip": {"message": "=inputs.messages.depSkip"},
                    },
                ],
                "locals": {
                    "mapKey": "a-key",
                },
                "return": {
                    "simple_cel": "=inputs.ints.a + inputs.ints.b",
                    "list": ["=inputs.ints.a + inputs.ints.b", 17, "constant"],
                    "map": {"mapKey": "=inputs.ints.a + inputs.ints.b"},
                },
            },
        )

        function, _ = prepared
        base_inputs = {
            "preconditions": {
                "skip": False,
                "permFail": False,
                "depSkip": False,
                "retry": False,
                "ok": False,
            },
            "messages": {
                "skip": "skip message",
                "permFail": "permFail message",
                "depSkip": "depSkip message",
            },
            "ints": {"a": 1, "b": 8},
        }

        reconcile_result = await reconcile.reconcile_value_function(
            location="test-fn",
            function=function,
            inputs=celpy.json_to_cel(base_inputs),
        )

        expected_value = {
            "simple_cel": 9,
            "list": [9, 17, "constant"],
            "map": {"mapKey": 9},
        }

        self.assertDictEqual(reconcile_result, expected_value)

    async def test_precondition_exits(self):
        predicate_pairs = (
            (
                "skip",
                result.Skip,
                {
                    "assert": "=!inputs.skip",
                    "skip": {"message": "skip message"},
                },
            ),
            (
                "permFail",
                result.PermFail,
                {
                    "assert": "=!inputs.permFail",
                    "permFail": {"message": "permFail message"},
                },
            ),
            (
                "depSkip",
                result.DepSkip,
                {
                    "assert": "=!inputs.depSkip",
                    "depSkip": {"message": "depSkip message"},
                },
            ),
            (
                "retry",
                result.Retry,
                {
                    "assert": "=!inputs.retry",
                    "retry": {"message": "retry message", "delay": 17},
                },
            ),
            (
                "ok",
                None,
                {"assert": "=!inputs.ok", "ok": {}},
            ),
        )

        predicates = [predicate for _, __, predicate in predicate_pairs]
        prepared_function = await prepare.prepare_value_function(
            cache_key="test", spec={"preconditions": predicates}
        )
        assert isinstance(prepared_function, tuple)
        function, _ = prepared_function

        base_inputs = {
            "skip": False,
            "permFail": False,
            "depSkip": False,
            "retry": False,
            "ok": False,
            "bogus": False,
        }
        for input_key, expected_type, _ in predicate_pairs:
            test_inputs = copy.deepcopy(base_inputs)
            test_inputs[input_key] = True

            reconcile_result = await reconcile.reconcile_value_function(
                location="test-fn",
                function=function,
                inputs=celpy.json_to_cel(test_inputs),
            )

            if expected_type is None:
                self.assertIsNone(reconcile_result)
            else:
                self.assertIsInstance(reconcile_result, expected_type)
                if input_key == "bogus":
                    self.assertTrue(
                        reconcile_result.message.startswith("Unknown predicate type")
                    )
                else:
                    self.assertEqual(reconcile_result.message, f"{input_key} message")

    async def test_simple_return(self):
        prepared = await prepare.prepare_value_function(
            cache_key="test",
            spec={
                "return": {
                    "value": "=inputs.a + inputs.b",
                    "list": ["=inputs.a + inputs.b", 17, "constant"],
                    "map": {"mapKey": "=inputs.a + inputs.b"},
                },
            },
        )

        function, _ = prepared

        base_inputs = {"a": 1, "b": 8}

        reconcile_result = await reconcile.reconcile_value_function(
            location="test-fn",
            function=function,
            inputs=celpy.json_to_cel(base_inputs),
        )

        expected_value = {
            "value": 9,
            "list": [9, 17, "constant"],
            "map": {"mapKey": 9},
        }

        self.assertDictEqual(reconcile_result, expected_value)

    async def test_return_with_locals(self):
        prepared = await prepare.prepare_value_function(
            cache_key="test",
            spec={
                "locals": {
                    "value": "=inputs.a + inputs.b",
                    "list": ["=inputs.a + inputs.b", 17, "constant"],
                    "map": {"mapKey": "=inputs.a + inputs.b"},
                },
                "return": {
                    "value": "=locals.value * locals.value",
                    "list": "=locals.list.map(value, string(value) + ' value')",
                    "map": "=locals.map.map(key, locals.map[key] * 3)",
                },
            },
        )

        function, _ = prepared

        base_inputs = {"a": 1, "b": 8}

        reconcile_result = await reconcile.reconcile_value_function(
            location="test-fn",
            function=function,
            inputs=celpy.json_to_cel(base_inputs),
        )

        expected_value = {
            "value": 81,
            "list": ["9 value", "17 value", "constant value"],
            "map": [27],
        }

        self.maxDiff = None
        self.assertDictEqual(reconcile_result, expected_value)

    async def test_corrupt_precondition(self):
        prepared = await prepare.prepare_value_function(
            cache_key="test",
            spec={
                "preconditions": [
                    {
                        "assert": "='a' + 9",
                        "skip": {"message": "skip message"},
                    },
                ],
            },
        )

        function, _ = prepared

        base_inputs = {}

        reconcile_result = await reconcile.reconcile_value_function(
            location="test-fn",
            function=function,
            inputs=celpy.json_to_cel(base_inputs),
        )

        self.assertIsInstance(reconcile_result, result.PermFail)
        self.assertIn("spec.preconditions", reconcile_result.message)

    async def test_corrupt_locals(self):
        prepared = await prepare.prepare_value_function(
            cache_key="test",
            spec={
                "locals": {
                    "busted": "='a' + 9",
                },
                "return": {
                    "value": "unused",  # Needed to prevent early-exit  eval of locals
                },
            },
        )

        function, _ = prepared

        base_inputs = {}

        reconcile_result = await reconcile.reconcile_value_function(
            location="test-fn",
            function=function,
            inputs=celpy.json_to_cel(base_inputs),
        )

        self.assertIsInstance(reconcile_result, result.PermFail)
        self.assertIn("spec.locals", reconcile_result.message)

    async def test_corrupt_return(self):
        prepared = await prepare.prepare_value_function(
            cache_key="test",
            spec={
                "return": {
                    "value": "='a' + 18",
                },
            },
        )

        function, _ = prepared

        base_inputs = {}

        reconcile_result = await reconcile.reconcile_value_function(
            location="test-fn",
            function=function,
            inputs=celpy.json_to_cel(base_inputs),
        )

        self.assertIsInstance(reconcile_result, result.PermFail)
        self.assertIn("spec.return", reconcile_result.message)



from collections import defaultdict
from typing import NamedTuple, Sequence
import asyncio
import time
import logging

logger = logging.getLogger(name="koreo.registry")


class Resource[T](NamedTuple):
    resource_type: type[T]
    name: str
    namespace: str | None = None


class Kill: ...


class ResourceEvent[T](NamedTuple):
    resource: Resource[T]
    event_time: float


type RegistryQueue = asyncio.Queue[ResourceEvent | Kill]


def register[T](
    registerer: Resource[T],
    queue: RegistryQueue | None = None,
) -> RegistryQueue:
    if registerer in _SUBSCRIPTION_QUEUES:
        return _SUBSCRIPTION_QUEUES[registerer]

    if not queue:
        queue = asyncio.LifoQueue[ResourceEvent | Kill]()

    _SUBSCRIPTION_QUEUES[registerer] = queue

    event_time = time.monotonic()
    notify_subscribers(notifier=registerer, event_time=event_time)

    logger.debug(f"Registering {registerer}")

    return queue


class SubscriptionCycle(Exception): ...


def subscribe(subscriber: Resource, resource: Resource):
    _check_for_cycles(subscriber, (resource,))

    _RESOURCE_SUBSCRIBERS[resource].add(subscriber)
    _SUBSCRIBER_RESOURCES[subscriber].add(resource)

    logger.debug(f"{subscriber} subscribing to {resource}")


def subscribe_only_to(subscriber: Resource, resources: Sequence[Resource]):
    _check_for_cycles(subscriber, resources)

    current = _SUBSCRIBER_RESOURCES[subscriber]

    new = set(resources)

    for resource in new - current:
        _RESOURCE_SUBSCRIBERS[resource].add(subscriber)

    for resource in current - new:
        _RESOURCE_SUBSCRIBERS[resource].remove(subscriber)

    _SUBSCRIBER_RESOURCES[subscriber] = new

    logger.debug(f"{subscriber} subscribing to {resources}")


def unsubscribe(unsubscriber: Resource, resource: Resource):
    _RESOURCE_SUBSCRIBERS[resource].remove(unsubscriber)
    _SUBSCRIBER_RESOURCES[unsubscriber].remove(resource)


def notify_subscribers(notifier: Resource, event_time: float):
    subscribers = _RESOURCE_SUBSCRIBERS[notifier]
    if not subscribers:
        logger.debug(f"{notifier} has no subscribers")
        return

    active_subscribers = [
        _SUBSCRIPTION_QUEUES[subscriber]
        for subscriber in subscribers
        if subscriber in _SUBSCRIPTION_QUEUES
    ]

    if not active_subscribers:
        logger.debug(f"{notifier} has no active subscribers")
        return

    logger.debug(f"{notifier}:{event_time} notifying to {subscribers}")

    for subscriber in active_subscribers:
        try:
            subscriber.put_nowait(
                ResourceEvent(resource=notifier, event_time=event_time)
            )
        except asyncio.QueueFull:
            pass
            # TODO: I think there is a way to monitor for stalled subscribers
            # then notify a house-keeper process to deal with it.

            # health_check_task = asyncio.create_task()
            # _CHECK_SUBSCRIBER_HEALTH.add(health_check_task)
            # health_check_task.add_done_callback(_CHECK_SUBSCRIBER_HEALTH.discard)


def get_subscribers(resource: Resource):
    return _RESOURCE_SUBSCRIBERS[resource]


def get_subscriptions(resource: Resource):
    return _SUBSCRIBER_RESOURCES[resource]


def kill_resource(resource: Resource) -> RegistryQueue | None:
    if resource not in _SUBSCRIPTION_QUEUES:
        return None

    _kill_resource(resource)


def deregister(deregisterer: Resource, deregistered_at: float):
    # This resource is no longer following any resources.
    subscribe_only_to(subscriber=deregisterer, resources=[])

    # Remove this resource's subscription queue
    if deregisterer in _SUBSCRIPTION_QUEUES:
        queue = _kill_resource(resource=deregisterer)
        assert queue  # Just for the type-checker

        del _SUBSCRIPTION_QUEUES[deregisterer]

        # This is to prevent blocking anything waiting for this resource to do
        # something.
        while not queue.empty():
            try:
                queue.get_nowait()
                queue.task_done()
            except asyncio.QueueEmpty:
                break

    # Inform subscribers of a change
    notify_subscribers(notifier=deregisterer, event_time=deregistered_at)


def _kill_resource(resource: Resource) -> RegistryQueue | None:
    queue = _SUBSCRIPTION_QUEUES[resource]
    try:
        queue.put_nowait(Kill())

    except asyncio.QueueShutDown:
        return queue

    except asyncio.QueueFull:
        pass

    queue.shutdown()

    return queue


def _check_for_cycles(subscriber: Resource, resources: Sequence[Resource]):
    # Simple, inefficient cycle detection. This is a simple brute-force check,
    # which hopefully given the problem space is sufficient.
    to_check: set[Resource] = set(resources)
    while to_check:
        if subscriber in to_check:
            raise SubscriptionCycle(f"Detected subscription cycle due to {subscriber}")

        next_check_set = set[Resource]()
        for check_resource in to_check:
            if check_resource not in _SUBSCRIBER_RESOURCES:
                continue

            next_check_set.update(_SUBSCRIBER_RESOURCES[check_resource])
        to_check = next_check_set


_RESOURCE_SUBSCRIBERS: defaultdict[Resource, set[Resource]] = defaultdict(set[Resource])
_SUBSCRIBER_RESOURCES: defaultdict[Resource, set[Resource]] = defaultdict(set[Resource])
_SUBSCRIPTION_QUEUES: dict[Resource, RegistryQueue] = {}


def _reset_registries():
    _RESOURCE_SUBSCRIBERS.clear()
    _SUBSCRIBER_RESOURCES.clear()

    for queue in _SUBSCRIPTION_QUEUES.values():
        try:
            queue.put_nowait(Kill())
        except (asyncio.QueueFull, asyncio.QueueShutDown):
            pass

        try:
            while not queue.empty():
                queue.get_nowait()
                queue.task_done()
        except asyncio.QueueEmpty:
            pass

    _SUBSCRIPTION_QUEUES.clear()



import copy
from datetime import UTC, datetime
from typing import TypedDict, NotRequired


class Condition(TypedDict):
    lastTransitionTime: NotRequired[str]
    lastUpdateTime: NotRequired[str]
    location: NotRequired[str | None]
    message: NotRequired[str | None]
    status: str
    reason: str
    type: str


Conditions = list[Condition]


def update_condition(conditions: Conditions, condition: Condition):
    conditions = copy.deepcopy(conditions)

    for index, candidate in enumerate(conditions):
        if candidate.get("type") == condition.get("type"):
            conditions[index] = _merge_conditions(candidate, condition)
            return conditions

    conditions.append(_new_condition(condition))
    return conditions


def _new_condition(new: Condition) -> Condition:
    event_time = datetime.now(UTC).isoformat()
    return Condition(
        lastUpdateTime=event_time,
        lastTransitionTime=event_time,
        type=new.get("type"),
        status=new.get("status"),
        reason=new.get("reason"),
        message=new.get("message"),
        location=new.get("location"),
    )


def _merge_conditions(base: Condition, updated: Condition) -> Condition:
    lastTransitionTime = base.get("lastTransitionTime")
    if base.get("status") != updated.get("status") or not lastTransitionTime:
        lastTransitionTime = datetime.now(UTC).isoformat()

    return Condition(
        type=base.get("type"),
        lastUpdateTime=datetime.now(UTC).isoformat(),
        lastTransitionTime=lastTransitionTime,
        status=updated.get("status"),
        reason=updated.get("reason"),
        message=updated.get("message"),
        location=updated.get("location"),
    )



from unittest.mock import AsyncMock
import asyncio
import unittest

import kr8s.asyncio

from koreo.resource_function.reconcile.kind_lookup import (
    get_plural_kind,
    _reset,
    _lookup_locks,
)


class TestGetFullKind(unittest.IsolatedAsyncioTestCase):
    def tearDown(self):
        _reset()

    async def test_ok_lookup(self):
        plural_kind = "unittesties"

        api_mock = AsyncMock(kr8s.asyncio.Api)
        api_mock.lookup_kind.return_value = (None, plural_kind, None)

        api_version = "unit.test/v1"

        result = await get_plural_kind(api_mock, "UnitTest", api_version)

        self.assertEqual(result, plural_kind)
        self.assertEqual(1, api_mock.lookup_kind.call_count)

    async def test_successive_lookups(self):
        plural_kind = "unittesties"

        api_mock = AsyncMock(kr8s.asyncio.Api)
        api_mock.lookup_kind.return_value = (None, plural_kind, None)

        api_version = "unit.test/v1"

        result = await get_plural_kind(api_mock, "UnitTest", api_version)
        self.assertEqual(result, plural_kind)

        result = await get_plural_kind(api_mock, "UnitTest", api_version)
        self.assertEqual(result, plural_kind)

        self.assertEqual(1, api_mock.lookup_kind.call_count)

    async def test_multiple_requests(self):
        plural_kind = "unittesties"

        async def lookup(_):
            await asyncio.sleep(0)
            return (None, plural_kind, None)

        api_mock = AsyncMock(kr8s.asyncio.Api)
        api_mock.lookup_kind.side_effect = lookup

        api_version = "unit.test/v1"

        tasks = [
            asyncio.create_task(get_plural_kind(api_mock, "UnitTest", api_version)),
            asyncio.create_task(get_plural_kind(api_mock, "UnitTest", api_version)),
            asyncio.create_task(get_plural_kind(api_mock, "UnitTest", api_version)),
        ]
        done, pending = await asyncio.wait(tasks)

        self.assertEqual(len(tasks), len(done))
        self.assertEqual(0, len(pending))

        for task in tasks:
            self.assertEqual(task.result(), plural_kind)

        self.assertEqual(1, api_mock.lookup_kind.call_count)

    async def test_missing_kind(self):
        api_mock = AsyncMock(kr8s.asyncio.Api)
        api_mock.lookup_kind.side_effect = ValueError("Kind not found")

        api_version = "unit.test/v1"

        plural_kind = await get_plural_kind(api_mock, "UnitTest", api_version)

        self.assertIsNone(plural_kind)

    async def test_2_timeout_retries(self):
        plural_kind = "unittesties"

        api_mock = AsyncMock(kr8s.asyncio.Api)
        api_mock.lookup_kind.side_effect = [
            asyncio.TimeoutError(),
            (None, plural_kind, None),
        ]

        api_version = "unit.test/v1"

        result = await get_plural_kind(api_mock, "UnitTest", api_version)
        self.assertEqual(result, plural_kind)

    async def test_too_many_timeout_retries(self):
        api_mock = AsyncMock(kr8s.asyncio.Api)
        api_mock.lookup_kind.side_effect = asyncio.TimeoutError()

        api_version = "unit.test/v1"

        with self.assertRaises(Exception):
            await get_plural_kind(api_mock, "unittest", api_version)

    async def test_lock_timeout(self):
        api_mock = AsyncMock(kr8s.asyncio.Api)
        api_mock.lookup_kind.side_effect = asyncio.TimeoutError()

        api_version = "unit.test/v1"

        lock_mock = AsyncMock(asyncio.Event)
        lock_mock.wait.side_effect = asyncio.TimeoutError()

        lookup_kind = f"unittest.{api_version}"
        _lookup_locks[lookup_kind] = lock_mock

        with self.assertRaises(Exception):
            await get_plural_kind(api_mock, "unittest", api_version)

    async def test_lock_cleared_on_errors(self):
        api_mock = AsyncMock(kr8s.asyncio.Api)
        api_mock.lookup_kind.side_effect = ZeroDivisionError("Unit Test")

        api_version = "unit.test/v1"

        with self.assertRaises(ZeroDivisionError):
            await get_plural_kind(api_mock, "unittest", api_version)



from importlib import resources
import logging
from pathlib import Path
from typing import Any

logger = logging.getLogger("koreo.schema")

import fastjsonschema
import yaml

from koreo.constants import DEFAULT_API_VERSION
from koreo.function_test.structure import FunctionTest
from koreo.resource_function.structure import ResourceFunction
from koreo.resource_template.structure import ResourceTemplate
from koreo.result import PermFail
from koreo.value_function.structure import ValueFunction
from koreo.workflow.structure import Workflow


NAME_CRD_MAP = {
    "functiontests.koreo.dev": FunctionTest,
    "resourcefunctions.koreo.dev": ResourceFunction,
    "resourcetemplates.koreo.dev": ResourceTemplate,
    "valuefunctions.koreo.dev": ValueFunction,
    "workflows.koreo.dev": Workflow,
}

_SCHEMA_VALIDATORS = {}


def validate(
    resource_type: type,
    spec: Any,
    schema_version: str | None = None,
    validation_required: bool = False,
):
    schema_validator = _get_validator(
        resource_type=resource_type, version=schema_version
    )
    if not schema_validator:
        if not validation_required:
            return None

        return PermFail(
            f"Schema validator not found for {resource_type.__name__} version {schema_version or DEFAULT_API_VERSION}",
        )

    try:
        schema_validator(spec)
    except fastjsonschema.JsonSchemaValueException as err:
        # This is hacky, and likely buggy, but it makes the messages easier to grok.
        validation_err = f"{err.rule_definition} {err}".replace(
            "data.", "spec."
        ).replace("data ", "spec ")
        return PermFail(validation_err)

    return None


def _get_validator(resource_type: type, version: str | None = None):
    if not _SCHEMA_VALIDATORS:
        load_bundled_schemas()

    if not version:
        version = DEFAULT_API_VERSION

    resource_version_key = f"{resource_type.__qualname__}:{version}"

    return _SCHEMA_VALIDATORS.get(resource_version_key)


def load_validator(resource_type_name: str, resource_schema: dict):
    spec = resource_schema.get("spec")
    if not spec:
        return None

    spec_names = spec.get("names")
    if spec_names:
        spec_kind = spec_names.get("kind", "<missing kind>")
    else:
        spec_kind = "<missing kind>"

    schema_specs = spec.get("versions")
    if not schema_specs:
        return None

    for schema_spec in schema_specs:
        version = schema_spec.get("name")
        if not version:
            continue

        schema_block = schema_spec.get("schema")
        if not schema_block:
            continue

        openapi_schema = schema_block.get("openAPIV3Schema")
        if not openapi_schema:
            continue

        openapi_properties = openapi_schema.get("properties")
        if not openapi_properties:
            continue

        openapi_spec = openapi_properties.get("spec")

        try:
            version_validator = fastjsonschema.compile(openapi_spec)
        except fastjsonschema.JsonSchemaDefinitionException:
            logger.exception(f"Failed to process {spec_kind} {version}")
            continue
        except AttributeError as err:
            logger.error(
                f"Probably encountered an empty `properties` block for {spec_kind} {version} (err: {err})"
            )
            raise

        resource_version_key = f"{resource_type_name}:{version}"
        _SCHEMA_VALIDATORS[resource_version_key] = version_validator


def load_validators_from_path(from_path: Path, clear_existing: bool = False):
    if clear_existing:
        _SCHEMA_VALIDATORS.clear()

    for resource in from_path.glob("*.yaml"):
        with resource.open() as crd_content:
            parsed = yaml.load_all(crd_content, Loader=yaml.SafeLoader)

            if not parsed:
                continue

            for chunk in parsed:
                chunk_kind = chunk.get("kind")
                if not chunk_kind or chunk_kind != "CustomResourceDefinition":
                    continue

                chunk_metadata = chunk.get("metadata")
                if not chunk_metadata:
                    continue

                chunk_resource = NAME_CRD_MAP.get(chunk_metadata.get("name"))
                if not chunk_resource:
                    continue

                load_validator(
                    resource_type_name=chunk_resource.__qualname__,
                    resource_schema=chunk,
                )


def load_bundled_schemas(clear_existing: bool = False):
    if clear_existing:
        _SCHEMA_VALIDATORS.clear()

    schema_module_resources = resources.path("koreo", "schema")

    with schema_module_resources as schema_path:
        load_validators_from_path(from_path=schema_path, clear_existing=clear_existing)



from typing import NamedTuple, Sequence
import copy
import json
import logging

logger = logging.getLogger("koreo.function")

from kr8s._objects import APIObject
import kr8s

import celpy
from celpy import celtypes

from koreo import cache
from koreo.cel import functions
from koreo.cel.encoder import convert_bools
from koreo.cel.evaluation import (
    evaluate,
    evaluate_overlay,
    evaluate_predicates,
    check_for_celevalerror,
)
from koreo.constants import (
    DEFAULT_LOAD_RETRY_DELAY,
    KOREO_DIRECTIVE_KEYS,
    LAST_APPLIED_ANNOTATION,
    PLURAL_LOOKUP_NEEDED,
)
from koreo.resource_template.structure import ResourceTemplate
from koreo.result import PermFail, Retry, UnwrappedOutcome, is_unwrapped_ok
from koreo.value_function.reconcile import reconcile_value_function

from koreo.resource_function import structure

from .kind_lookup import get_plural_kind
from .validate import validate_match


class Result(NamedTuple):
    outcome: UnwrappedOutcome[celtypes.Value]
    resource_id: dict | None = None


async def reconcile_resource_function(
    api: kr8s.Api,
    location: str,
    function: structure.ResourceFunction,
    owner: tuple[str, dict],
    inputs: celtypes.Value,
) -> Result:
    full_inputs: dict[str, celtypes.Value] = {
        "inputs": inputs,
    }

    if precondition_error := evaluate_predicates(
        predicates=function.preconditions,
        inputs=full_inputs,
        location=f"{location}:spec.preconditions",
    ):
        return Result(outcome=precondition_error)

    match evaluate(
        expression=function.local_values,
        inputs=full_inputs,
        location=f"{location}:spec.locals",
    ):
        case PermFail() as err:
            return Result(outcome=err)
        case None:
            full_inputs["locals"] = celtypes.MapType({})
        case celtypes.MapType() as local_values:
            full_inputs["locals"] = local_values
        case bad_type:
            # Due to validation within `prepare`, this should never happen.
            return Result(
                outcome=PermFail(
                    message=f"Invalid `locals` expression type ({type(bad_type)})",
                    location=f"{location}:spec.locals",
                )
            )

    ###########################
    # Start Kubernetes Specific
    ###########################

    reconcile_result = await reconcile_krm_resource(
        api=api,
        crud_config=function.crud_config,
        owner=owner,
        inputs=full_inputs,
    )
    resource_id = reconcile_result.resource_id
    if resource_id:
        resource_id["resourceFunction"] = function.name

    if not is_unwrapped_ok(reconcile_result.result):
        return Result(outcome=reconcile_result.result, resource_id=resource_id)

    full_inputs["resource"] = celpy.json_to_cel(reconcile_result.result)

    #########################
    # End Kubernetes Specific
    #########################

    if postcondition_error := evaluate_predicates(
        predicates=function.postconditions,
        inputs=full_inputs,
        location=f"{location}:spec.postconditions",
    ):
        return Result(outcome=postcondition_error)

    return Result(
        outcome=evaluate(
            expression=function.return_value,
            inputs=full_inputs,
            location=f"{location}:spec.return",
        ),
        resource_id=resource_id,
    )


class ReconcileResult(NamedTuple):
    result: PermFail | Retry | dict
    resource_id: dict | None = None


async def reconcile_krm_resource(
    api: kr8s.Api,
    crud_config: structure.CRUDConfig,
    owner: tuple[str, dict],
    inputs: dict[str, celtypes.Value],
) -> ReconcileResult:
    match evaluate(
        expression=crud_config.resource_id,
        inputs=inputs,
        location="spec.apiConfig.name",
    ):
        case PermFail(message=message, location=name_location):
            return ReconcileResult(
                result=PermFail(
                    message=message,
                    location=(
                        name_location if name_location else f"spec.apiConfig.name"
                    ),
                )
            )

        case None:
            return ReconcileResult(
                result=PermFail(
                    message="Could not evaluate `spec.apiConfig.name`, evaluated to `null`",
                    location=f"spec.apiConfig.name",
                )
            )

        case celtypes.MapType({"name": name_value}) as resource_name_values:
            name = f"{name_value}"
            namespace_value = resource_name_values.get("namespace")
            if namespace_value:
                namespace = f"{namespace_value}"
            else:
                namespace = None

            if not namespace and crud_config.resource_api.namespaced:
                return ReconcileResult(
                    result=PermFail(
                        message="`namespace` is required when `spec.apiConfig.namespaced` is `true`",
                        location=f"spec.apiConfig.namespace",
                    )
                )

        case bad_type:
            # Due to validation within `prepare`, this should never happen.
            return ReconcileResult(
                result=PermFail(
                    message=f"Invalid `spec.apiConfig.name` expression type ({type(bad_type)})",
                    location=f"spec.apiConfig.name",
                )
            )

    # TODO: Would we rather do this at prepare time? Or, perhaps there should
    # be some other process that ensures we lookup the plurals in advance when
    # the resource function is loaded?
    if crud_config.resource_api.plural == PLURAL_LOOKUP_NEEDED:
        kind = crud_config.resource_api.kind

        plural = await get_plural_kind(
            api=api,
            kind=kind,
            api_version=crud_config.resource_api.version,
        )

        if not plural:
            message = (
                f"Dynamic kind-plural lookup failed for '{kind}'. Set plural "
                "explicitly in apiConfig."
            )
            logger.warning(message)
            return ReconcileResult(
                result=PermFail(message=message, location=f"spec.apiConfig.plural")
            )

        crud_config.resource_api.endpoint = plural
        crud_config.resource_api.plural = plural

    resource_id = {
        "apiVersion": crud_config.resource_api.version,
        "kind": crud_config.resource_api.kind,
        "plural": crud_config.resource_api.plural,
        "name": name,
        "readonly": crud_config.readonly,
    }
    if namespace:
        full_resource_name = f"{crud_config.resource_api.kind}:{namespace}:{name}"
        resource_id["namespace"] = namespace
    else:
        full_resource_name = f"{crud_config.resource_api.kind}:{name}"

    api_resource = await load_api_resource(
        api=api,
        resource_api=crud_config.resource_api,
        name=name,
        namespace=namespace,
    )
    if not is_unwrapped_ok(api_resource):
        return ReconcileResult(result=api_resource, resource_id=resource_id)

    if crud_config.delete_if_exists:
        if not api_resource:
            return ReconcileResult(
                result={},
                resource_id=resource_id,
            )

        await api_resource.delete()
        return ReconcileResult(
            result=Retry(
                message=f"Deleting {full_resource_name}.",
                delay=DEFAULT_LOAD_RETRY_DELAY,
            ),
            resource_id=resource_id,
        )

    if not api_resource and (crud_config.readonly or not crud_config.create.enabled):
        # This is where we used virtual, but I think ValueFunction replaces that?
        # TODO: Return None or return base?
        return ReconcileResult(
            result=Retry(
                message=f"{full_resource_name} not found. Waiting...",
                delay=DEFAULT_LOAD_RETRY_DELAY,
            ),
            resource_id=resource_id,
        )

    if api_resource and crud_config.readonly:
        return ReconcileResult(result=api_resource.raw, resource_id=resource_id)

    forced_overlay = _forced_overlay(
        resource_api=crud_config.resource_api, name=name, namespace=namespace
    )
    if not is_unwrapped_ok(forced_overlay):
        return ReconcileResult(
            result=forced_overlay,
            resource_id=resource_id,
        )

    expected_resource = await _construct_resource_template(
        resource_template=crud_config.resource_template,
        inputs=inputs,
        forced_overlay=forced_overlay,
        full_resource_name=full_resource_name,
    )
    if not is_unwrapped_ok(expected_resource):
        return ReconcileResult(
            result=expected_resource,
            resource_id=resource_id,
        )

    if crud_config.overlays:
        if not is_unwrapped_ok(crud_config.overlays):
            return ReconcileResult(
                result=crud_config.overlays,
                resource_id=resource_id,
            )
        expected_resource = await _materialize_from_overlays(
            resource=expected_resource,
            overlay_steps=crud_config.overlays,
            inputs=inputs,
            forced_overlay=forced_overlay,
            full_resource_name=full_resource_name,
        )
        if not is_unwrapped_ok(expected_resource):
            return ReconcileResult(
                result=expected_resource,
                resource_id=resource_id,
            )

    if not api_resource:
        return ReconcileResult(
            result=await _create_api_resource(
                api=api,
                resource_api=crud_config.resource_api,
                create=crud_config.create,
                namespace=namespace,
                owned_resource=crud_config.own_resource,
                owner=owner,
                inputs=inputs,
                resource_view=expected_resource,
                forced_overlay=forced_overlay,
                full_resource_name=full_resource_name,
            ),
            resource_id=resource_id,
        )

    owner_namespace, owner_ref = owner
    should_own = crud_config.own_resource and owner_namespace == namespace
    if should_own:
        owner_reffed = _validate_owner_reffed(api_resource.raw, owner_ref)
    else:
        owner_reffed = True

    converted_resource = convert_bools(expected_resource)

    last_applied = _extract_last_applied(api_resource.raw)

    resource_match = validate_match(
        target=converted_resource,
        actual=api_resource.raw,
        last_applied_value=last_applied,
    )
    if resource_match.match and owner_reffed:
        logger.debug(f"{full_resource_name} matched spec, no update required.")

        return ReconcileResult(result=api_resource.raw, resource_id=resource_id)

    match crud_config.update:
        case structure.UpdateNever():
            logger.debug(
                f"Skipping update (behavior.update is 'never') for {full_resource_name}."
            )
            return ReconcileResult(result=api_resource.raw, resource_id=resource_id)

        case structure.UpdateRecreate(delay=delay):
            await api_resource.delete()
            return ReconcileResult(
                result=Retry(
                    message=(
                        f"Deleting {full_resource_name} to recreate due to "
                        f"{' -> '.join(resource_match.differences)}."
                    ),
                    delay=delay,
                    location="spec.update.recreate",
                ),
                resource_id=resource_id,
            )

        case structure.UpdatePatch(delay=delay):
            if should_own and not owner_reffed:
                owner_refs = _updated_owner_refs(api_resource.raw, owner_ref)
                if not is_unwrapped_ok(owner_refs):
                    return ReconcileResult(result=owner_refs, resource_id=resource_id)

                converted_resource["metadata"]["ownerReferences"] = owner_refs

            await api_resource.patch(_prepare_for_api(converted_resource))
            return ReconcileResult(
                result=Retry(
                    message=(
                        f"Patching `{full_resource_name}` due to "
                        f"{' -> '.join(resource_match.differences)}."
                    ),
                    delay=delay,
                    location="spec.update.patch",
                ),
                resource_id=resource_id,
            )


async def load_api_resource(
    api: kr8s.Api,
    resource_api: type[APIObject],
    name: str,
    namespace: str | None,
):
    try:
        matches = [
            match
            async for match in api.async_get(
                resource_api,
                name,
                namespace=namespace,
            )
        ]
    except kr8s.NotFoundError:
        matches = None
    except kr8s.ServerError as err:
        if not err.response or err.response.status_code != 404:
            msg = (
                f"ServerError loading {resource_api.kind} {name}. ({type(err)}: {err})"
            )
            logger.exception(msg)

            return Retry(
                message=msg, delay=DEFAULT_LOAD_RETRY_DELAY, location="load resource"
            )

        matches = None

    except Exception as err:
        msg = f"Failure loading {resource_api.kind} {name}. ({type(err)}: {err})"
        logger.exception(msg)
        return Retry(
            message=msg, delay=DEFAULT_LOAD_RETRY_DELAY, location="load resource"
        )

    match matches:
        case list() if len(matches) == 1:
            return matches[0]

        case APIObject():
            return matches

        case list() if len(matches) > 1:
            return Retry(
                message=f"{resource_api.kind}/{name} resource matched multiple resources.",
                delay=DEFAULT_LOAD_RETRY_DELAY,
            )

        case list():
            return None


async def _construct_resource_template(
    inputs: dict[str, celtypes.Value],
    resource_template: (
        structure.InlineResourceTemplate | structure.ResourceTemplateRef | None
    ),
    forced_overlay: celtypes.MapType,
    full_resource_name: str,
):
    match resource_template:
        case None:
            return forced_overlay

        case structure.ResourceTemplateRef(name=template_name):
            match evaluate(
                expression=template_name,
                inputs=inputs,
                location=f"{full_resource_name}:spec.resourceTemplateRef.name<eval>",
            ):
                case PermFail() as err:
                    return err
                case (celtypes.StringType() | str()) as template_key:
                    pass
                case _ as bad_type:
                    return PermFail(
                        message=f"Expected string, but received '{bad_type}' ({type(bad_type)}) at `spec.resourceTemplateRef.name`",
                        location=f"{full_resource_name}:spec.resourceTemplateRef.name<eval>",
                    )

            dynamic_resource_template = cache.get_resource_from_cache(
                resource_class=ResourceTemplate, cache_key=template_key
            )

            if not dynamic_resource_template:
                return Retry(
                    message=f"ResourceTemplate:{template_key} not found.",
                    location=f"{full_resource_name}:spec.resourceTemplateRef.name<load>",
                    delay=DEFAULT_LOAD_RETRY_DELAY,
                )

            if not is_unwrapped_ok(dynamic_resource_template):
                return Retry(
                    message=f"ResourceTemplate:{template_key} is not ready ({dynamic_resource_template.message})",
                    location=f"{full_resource_name}:spec.resourceTemplateRef.name<load>",
                    delay=DEFAULT_LOAD_RETRY_DELAY,
                )

            materialized = dynamic_resource_template.template

        case structure.InlineResourceTemplate(template=template):
            match evaluate(
                expression=template,
                inputs=inputs,
                location=f"{full_resource_name}:spec.resource",
            ):
                case PermFail() as err:
                    return err
                case celtypes.MapType() as materialized:
                    pass
                case None:
                    materialized = celtypes.MapType()
                case _ as bad_type:
                    return PermFail(
                        message=f"Expected mapping, but received {type(bad_type)} for `spec.resource`",
                        location=f"{full_resource_name}:spec.resource",
                    )

    materialized = functions._overlay(resource=materialized, overlay=forced_overlay)
    if err := check_for_celevalerror(
        materialized, location="spec.resource<security overlay>"
    ):
        return err

    # This can not be, it is for type checkers
    assert not isinstance(materialized, celpy.CELEvalError)

    return materialized


async def _materialize_from_overlays(
    resource: celtypes.MapType,
    overlay_steps: Sequence[structure.InlineOverlay | structure.ValueFunctionOverlay],
    inputs: dict[str, celtypes.Value],
    forced_overlay: celtypes.MapType,
    full_resource_name: str,
):
    location = f"{full_resource_name}:spec.overlays"

    new_resource = resource
    for idx, overlay_step in enumerate(overlay_steps):
        if overlay_step.skip_if:
            match evaluate(
                overlay_step.skip_if,
                inputs=inputs,
                location=f"{location}[{idx}].skipIf",
            ):
                case PermFail() as err:
                    return err

                case celtypes.BoolType() as skip:
                    if skip:
                        continue

                case _ as bad_type:
                    return PermFail(
                        message=f"Expected boolean, but received {type(bad_type)} for 'spec.overlays[{idx}].skipIf'",
                        location=f"{location}[{idx}].skipIf",
                    )

        match overlay_step:
            case structure.InlineOverlay(overlay=overlay):
                match evaluate_overlay(
                    overlay=overlay,
                    inputs=inputs,
                    base=new_resource,
                    location=f"{location}[{idx}].overlay",
                ):
                    case PermFail() as err:
                        return err

                    case celtypes.MapType() as overlaid:
                        new_resource = overlaid

                    case _ as bad_type:
                        return PermFail(
                            message=f"Expected mapping, but received {type(bad_type)} for '{location}[{idx}].overlay'",
                            location=f"{location}[{idx}].overlay",
                        )

            case structure.ValueFunctionOverlay(
                inputs=overlay_inputs, overlay=overlay_function
            ):
                evaluated_inputs = evaluate(
                    expression=overlay_inputs,
                    inputs=inputs,
                    location=f"{location}[{idx}].inputs",
                )

                if not is_unwrapped_ok(evaluated_inputs):
                    return evaluated_inputs

                overlaid = await reconcile_value_function(
                    location=f"{location}[{idx}].overlayRef<eval>",
                    function=overlay_function,
                    inputs=evaluated_inputs,
                    value_base=new_resource,
                )
                if not is_unwrapped_ok(overlaid):
                    return overlaid

                if not isinstance(overlaid, celtypes.MapType):
                    return PermFail(
                        message=f"Expected mapping, but received {type(overlaid)} for '{location}[{idx}].overlayRef<eval>'",
                        location=f"{location}[{idx}].overlayRef<eval>",
                    )

                new_resource = overlaid

        if err := check_for_celevalerror(
            new_resource, location=f"{location}[{idx}]<apply>"
        ):
            return err

        assert not isinstance(new_resource, celpy.CELEvalError)

    new_resource = functions._overlay(resource=new_resource, overlay=forced_overlay)
    if err := check_for_celevalerror(
        new_resource, location="spec.overlays<security overlay>"
    ):
        return err

    return new_resource


async def _create_api_resource(
    api: kr8s.Api,
    resource_api: type[APIObject],
    namespace: str | None,
    create: structure.Create,
    owned_resource: bool,
    owner: tuple[str, dict],
    inputs: dict[str, celtypes.Value],
    resource_view: celtypes.MapType | None,
    forced_overlay: celtypes.MapType,
    full_resource_name: str,
):
    if not resource_view:
        resource_view = celtypes.MapType()

    if create.overlay:
        match evaluate_overlay(
            overlay=create.overlay,
            inputs=inputs,
            base=resource_view,
            location="spec.create.overlay",
        ):
            case PermFail() as failure:
                return failure
            case celtypes.MapType() as resource_view:
                # Just needed to update resource_view
                pass
            case _ as bad_type:
                return PermFail(
                    message=f"Expected mapping, but received {type(bad_type)} for 'spec.create.overlay'",
                    location="spec.create.overlay",
                )

        if err := check_for_celevalerror(resource_view, location="spec.create.overlay"):
            return err

        # Purely for TypeChecker; check_for_celevalerror eliminates these.
        assert not isinstance(resource_view, celpy.CELEvalError)

    match functions._overlay(resource=resource_view, overlay=forced_overlay):
        case celpy.CELEvalError() as err:
            if updated_err := check_for_celevalerror(
                err, location="spec.create.overlay(name apply)"
            ):
                return updated_err

            # This should never happen.
            return PermFail(
                message=f"Error applying name/namesace overlay: {err}",
                location="spec.create.overlay(apply)",
            )
        case celtypes.MapType() as resource_view:
            pass

    if err := check_for_celevalerror(resource_view, location="spec.create.overlay"):
        return err

    owner_namespace, owner_ref = owner
    if owned_resource and owner_namespace == namespace:
        owner_refs = _updated_owner_refs(resource_view, owner_ref)
        if not is_unwrapped_ok(owner_refs):
            return owner_refs

        resource_view["metadata"]["ownerReferences"] = owner_refs

    converted_resource = convert_bools(resource_view)

    if not isinstance(converted_resource, (celtypes.MapType, dict)):
        return PermFail(
            message=f"Unexpected issue with encoding resource, received {type(converted_resource)}",
            location="<api-create-convert>",
        )

    new_resource = resource_api(
        api=api,
        resource=_prepare_for_api(converted_resource),
        namespace=namespace,
    )

    try:
        await new_resource.create()
    except kr8s.ServerError as err:
        if err.response and err.response.status_code == 409:
            return Retry(
                message=f"Waiting on competing creation of {full_resource_name}.",
                delay=create.delay,
                location="spec.create(contention)",
            )

        # TODO: Review possible error codes and determine which should retry

        logger.error(f"K8s API Server error: {err.status}")
        return PermFail(
            message=f"Error creating {full_resource_name}: {err.status}",
            location="spec.create",
        )

    except Exception as err:
        # TODO: Probably need to catch timeouts and retry here.
        logger.exception(f"Unhandled error calling create: {err}")
        return PermFail(
            message=f"Error creating {full_resource_name}: {err}",
            location="spec.create",
        )

    return Retry(
        message=f"Creating {full_resource_name}.",
        delay=create.delay,
        location="spec.create",
    )


def _forced_overlay(resource_api: type[APIObject], name: str, namespace: str | None):
    # Only include namespace when the resource API is namespaced
    metadata: dict[str, str] = {"name": name}
    if namespace is not None:
        metadata["namespace"] = namespace

    forced_overlay = celpy.json_to_cel(
        {
            "apiVersion": resource_api.version,
            "kind": resource_api.kind,
            "metadata": metadata,
        }
    )

    # Perhaps this could occur with some corrupt name config?
    if not isinstance(forced_overlay, celtypes.MapType):
        return PermFail(
            message=f"Unexpected issue with kind/name-overlay, received {type(forced_overlay)}",
            location="<api-create-overlay>",
        )

    return forced_overlay


def _updated_owner_refs(resource_view: dict, owner_ref: dict):
    match resource_view:
        case {"metadata": metadata}:
            pass
        case _:
            return PermFail(
                message=f"Missing resource while adding owner ref",
                location="<api-add-owner>",
            )

    match metadata:
        case {"ownerReferences": owner_refs}:
            pass
        case dict():
            return [owner_ref]
        case _:
            return PermFail(
                message=f"Corrupt `metadata` while adding owner ref",
                location="<api-add-owner>",
            )

    if not owner_refs:
        return [owner_ref]

    if not isinstance(owner_refs, (list, tuple)):
        return PermFail(
            message=f"Corrupt `ownerReferences` while adding owner ref",
            location="<api-add-owner>",
        )

    trigger_uid = owner_ref.get("uid")

    for current_ref in owner_refs:
        if current_ref.get("uid") == trigger_uid:
            return owner_refs

    updated_owners = copy.deepcopy(owner_refs)
    updated_owners.append(owner_ref)
    return updated_owners


def _validate_owner_reffed(resource_view: dict, owner_ref: dict):
    match resource_view:
        case {"metadata": metadata}:
            pass
        case _:
            return PermFail(
                message=f"Missing resource while adding owner ref",
                location="<api-add-owner>",
            )

    match metadata:
        case {"ownerReferences": owner_refs}:
            pass
        case dict():
            return False
        case _:
            return PermFail(
                message=f"Corrupt `metadata` while adding owner ref",
                location="<api-add-owner>",
            )

    if not owner_refs:
        return False

    if not isinstance(owner_refs, (list, tuple)):
        return PermFail(
            message=f"Corrupt `ownerReferences` while adding owner ref",
            location="<api-add-owner>",
        )

    trigger_uid = owner_ref.get("uid")

    for current_ref in owner_refs:
        if current_ref.get("uid") == trigger_uid:
            return True

    return False


def _extract_last_applied(resource: dict) -> dict | None:
    if not resource:
        return None

    metadata = resource.get("metadata")
    if not metadata:
        return None

    annotations = metadata.get("annotations")
    if not annotations:
        return None

    last_applied = annotations.get(LAST_APPLIED_ANNOTATION)
    if not last_applied:
        return None

    return json.loads(last_applied)


def _prepare_for_api(obj: dict) -> dict:
    prepared = _strip_koreo_directives(obj)

    dumped = json.dumps(prepared)

    if "metadata" not in prepared:
        prepared["metadata"] = {}

    if "annotations" not in prepared["metadata"]:
        prepared["metadata"]["annotations"] = {}

    prepared["metadata"]["annotations"][LAST_APPLIED_ANNOTATION] = dumped

    return prepared


def _strip_koreo_directives[T](obj: T) -> T:
    match obj:
        case dict():
            return {
                key: _strip_koreo_directives(sub_obj)
                for key, sub_obj in obj.items()
                if key not in KOREO_DIRECTIVE_KEYS
            }

        case list() | set() | tuple():
            return [_strip_koreo_directives(sub_obj) for sub_obj in obj]

        case _:
            return obj
